# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines integrators which form the core of the solvers."""

from functools import partial
from typing import Callable, Any, Tuple, Dict, NamedTuple

import jax.numpy as jnp
import numpy as onp
from jax import random, tree_unflatten, tree_flatten, grad, lax, tree_util, \
  value_and_grad, named_call

from jax_sgmc.adaption import AdaptionStrategy, MassMatrix
from jax_sgmc.data import RandomBatch, CacheState
from jax_sgmc.potential import StochasticPotential
from jax_sgmc.scheduler import schedule
from jax_sgmc.util import Array, tree_scale, tree_add, tensor_matmul, \
  tree_dot, Tensor, tree_multiply

PyTree = Any


class LeapfrogState(NamedTuple):
  """State of the reversible and friction leapfrog integrator.

  Attributes:
    positions: Latent variables in hamiltonian dynamics formulation
    momentum: Momentum with the same shape as the latent variables
    potential: Accumulated energy to calculate MH-correction step
    model_state: State of the model in the likelihood
    data_state: State of the random data function
    key: PRNGKey
  """
  positions: PyTree
  momentum: PyTree
  potential: Array
  model_state: PyTree
  data_state: CacheState
  key: Array
  extra_fields: PyTree = None


class ObaboState(NamedTuple):
  """State of the OBABO integrator.

  Attributes:
    positions: Latent variables in hamiltonian dynamics formulation
    momentum: Momentum with the same shape as the latent variables
    potential: Stochastic potential
    model_state: State of the model in the likelihood
    data_state: State of the random data function
    key: PRNGKey
    kinetic_energy_start: Kinetic energy after the first 1/4-step
    kinetic_energy_end: Kinetic energy after the last 3/4-step
  """
  positions: PyTree
  momentum: PyTree
  potential: Array
  model_state: PyTree
  data_state: CacheState
  key: Array
  kinetic_energy_start: Array
  kinetic_energy_end: Array


class LangevinState(NamedTuple):
  """State of the langevin diffusion integrator.

  Attributes:
    latent_variables: Current latent variables
    key: PRNGKey
    adapt_state: Containing quantities such as momentum for adaption
    data_state: State of the reference data cache
    model_state: Variables not considered during inference
    potential: Stochastic potential from last evaluation
    variance: Variance of stochastic potential over mini-batch
  """
  latent_variables: PyTree
  model_state: PyTree
  key: Array
  adapt_state: Any
  data_state: CacheState
  potential: Array
  variance: Array


def init_mass(mass) -> MassMatrix:
  """Initializes a diagonal mass tensor.

  Args:
    mass: Diagonal mass which has the same tree structure as the sample.

  Returns:
    Returns a diagonal mass matrix.
  """
  inv_mass = tree_util.tree_map(
    lambda x: jnp.power(x, -1.0),
    mass)
  sqrt_mass = tree_util.tree_map(
    jnp.sqrt,
    mass)
  inv_mass = Tensor(ndim=1, tensor=inv_mass)
  sqrt_mass = Tensor(ndim=1, tensor=sqrt_mass)
  return MassMatrix(inv=inv_mass, sqrt=sqrt_mass)


def random_tree(key, a):
  """Build a tree shaped like a where all nodes are normally distributed.

  Arguments:
    key: PRNGKey
    a: PyTree defining the shape of the output

  Returns:
    Tree shaped like a with normal distributed leaves.

  """
  leaves, tree_def = tree_flatten(a)
  splits = random.split(key, len(leaves))
  noise_leaves = [random.normal(split, leaf.shape)
                  for split, leaf in zip(splits, leaves)]
  noise_tree = tree_unflatten(tree_def, noise_leaves)
  return noise_tree


def obabo(potential_fn: StochasticPotential,
          batch_fn: RandomBatch,
          steps: Array = 10,
          friction: Array = 1.0,
          const_mass: PyTree = None,
          ) -> Tuple[Callable, Callable, Callable]:
  """Initializes the OBABO integration scheme.

  The OBABO integration scheme is reversible even when using stochastic
  gradients and provides second order accuracy.

  [1] https://arxiv.org/abs/2102.01691

  Args:
    potential_fn: Likelihood and prior applied over a minibatch of data
    batch_fn: Function to draw a mini-batch of reference data
    steps: Number of integration steps.
    friction: Controls impact of momentum from previous step
    const_mass: Mass matrix if no matrix is adapted. Must have the same tree
      structure as the sample

  Returns:
    Returns a function running the time OBABO integrator for T
    steps.

  """

  init_data, get_data, _ = batch_fn
  stochastic_gradient = value_and_grad(potential_fn, argnums=0, has_aux=True)

  # Calculate the inverse and the square root
  if const_mass:
    const_mass = init_mass(const_mass)
  else:
    const_mass = None

  # Helper functions to calculate the kinetic energy, update the position,
  # refresh and update the momentum.

  def _kinetic_energy(mass, momentum):
    scaled_momentum = tensor_matmul(mass.inv, momentum)
    return 0.5 * tree_dot(momentum, scaled_momentum)

  def _position_update(scale, mass, position, momentum):
    scaled_momentum = tensor_matmul(mass.inv, momentum)
    scaled_momentum = tree_scale(scale, scaled_momentum)
    return tree_add(position, scaled_momentum)

  # Half step if scale == 0.5 * step_size
  def _momentum_update(scale, gradient, momentum):
    scaled_gradient = tree_scale(-1.0 * scale, gradient)
    return tree_add(scaled_gradient, momentum)

  # Add noise to momentum
  def _momentum_resampling(parameters, mass, momentum, split):
    noise = random_tree(split, momentum)
    scaled_noise = tensor_matmul(mass.sqrt, noise)
    permanence = jnp.exp(-friction * parameters.step_size)
    momentum_noise = tree_scale(
      jnp.sqrt((1 - permanence) * parameters.temperature),
      scaled_noise)
    decayed_momentum = tree_scale(jnp.sqrt(permanence), momentum)
    return tree_add(decayed_momentum, momentum_noise)

  # A single OBABO-step of the integrator
  def _leapfrog_steps(state: ObaboState,
                      step: jnp.array,
                      mass: PyTree,
                      parameters: schedule = None):
    del step
    key, split1, split2 = random.split(state.key, num=3)

    refreshed_momentum = _momentum_resampling(
      parameters,
      mass,
      state.momentum,
      split1)

    # The kinetic energy from the first step is necessary to calculate the
    # acceptance probability
    # start_energy = lax.select(
    #   step == 0,
    #   _kinetic_energy(mass, refreshed_momentum),
    #   state.kinetic_energy_start)
    start_energy = state.kinetic_energy_start + _kinetic_energy(mass, refreshed_momentum)

    # Momentum update with stochastic gradient
    data_state, mini_batch = get_data(state.data_state, information=True)
    (pot_before, model_state), gradient = stochastic_gradient(
      state.positions,
      mini_batch,
      state=state.model_state)
    first_updated_momentum = _momentum_update(
      0.5 * parameters.step_size,
      gradient,
      refreshed_momentum)

    # Position update with momentum
    updated_positions = _position_update(
      parameters.step_size,
      mass,
      state.positions,
      first_updated_momentum)

    # Momentum update with stochastic gradient
    data_state, mini_batch = get_data(data_state, information=True)
    (pot_after, model_state), gradient = stochastic_gradient(
      updated_positions,
      mini_batch,
      state=model_state )
    second_updated_momentum = _momentum_update(
      0.5 * parameters.step_size,
      gradient,
      first_updated_momentum)

    final_refreshed_momentum = _momentum_resampling(
      parameters,
      mass,
      second_updated_momentum,
      split2)

    # The kinetic energy of the last step is necessary to
    # calculate the acceptance probability for the MH step.
    end_energy = state.kinetic_energy_end + _kinetic_energy(mass, second_updated_momentum)

    new_state = ObaboState(
      potential=0.5 * (pot_before + pot_after),
      positions=updated_positions,
      momentum=final_refreshed_momentum,
      key=key,
      data_state=data_state,
      model_state=model_state,
      kinetic_energy_start=start_energy,
      kinetic_energy_end=end_energy)

    return new_state, None

  def init_fn(init_sample: PyTree,
              key: Array = None,
              batch_kwargs: Dict = None,
              init_model_state: PyTree = None):
    """Initializes the initial state of the integrator.

    Args:
      init_sample: Initial latent variables
      key: Initial PRNGKey
      batch_kwargs: Determine the initial state of the random data chain
      init_model_state: State of the model.
      mass: Mass matrix

    Returns:
      Returns the initial state of the integrator.

    """

    # Initializing the initial state here makes it easier to add additional
    # variables which might be only necessary in special case
    if batch_kwargs is None:
      batch_kwargs = {}

    reference_data_state = init_data(**batch_kwargs)

    if key is None:
      key = random.PRNGKey(0)

    momentum = tree_util.tree_map(jnp.zeros_like, init_sample)

    init_state = ObaboState(
      kinetic_energy_start=jnp.array(0.0),
      kinetic_energy_end=jnp.array(0.0),
      potential=jnp.array(0.0),
      key=key,
      positions=init_sample,
      momentum=momentum,
      data_state=reference_data_state,
      model_state=init_model_state)

    return init_state

  @partial(named_call, name='obabo_integration')
  def integrate(state: ObaboState,
                parameters: schedule,
                mass: PyTree = None
                ) -> ObaboState:

    # If the mass is not adapted, take the constant mass if provided
    if const_mass is None:
      cms = init_mass(tree_util.tree_map(jnp.ones_like, state.positions))
    else:
      cms = const_mass

    if mass is None:
      mass = cms

    # Leapfrog integration
    state, _ = lax.scan(
      partial(_leapfrog_steps, parameters=parameters, mass=mass),
      state,
      onp.arange(steps))

    return state

  def get_fn(state: LeapfrogState) -> Dict[str, PyTree]:
    """Returns the latent variables."""
    return {"variables": state.positions,
            "energy": state.potential,
            "model_state": state.model_state}

  return init_fn, integrate, get_fn


def reversible_leapfrog(potential_fn: StochasticPotential,
                        batch_fn: RandomBatch,
                        steps: int = 10,
                        friction: [float, Array] = 0.25,
                        const_mass: PyTree = None
                        ) -> Tuple[Callable, Callable, Callable]:
  """Initializes a reversible leapfrog integrator.

  AMAGOLD requires a reversible leapfrog integrator with half step at the
  beginning and end.

  Args:
    potential_fn: Likelihood and prior applied over a minibatch of data
    batch_fn: Function to draw a mini-batch of reference data
    steps: Number of intermediate leapfrog steps
    friction: Decay of momentum to counteract induced noise due to stochastic
     gradients
    const_mass: Mass matrix to be used when no mass matrix is adapted

  Returns:
    Returns a function running the time reversible leapfrog integrator for T
    steps.

  """

  init_data, get_data, _ = batch_fn
  stochastic_gradient = grad(potential_fn, has_aux=True)

  # Calculate the inverse and the square root
  if const_mass:
    const_mass = init_mass(const_mass)
  else:
    const_mass = None

  def _position_update(scale, mass, position, momentum):
    scaled_momentum = tensor_matmul(mass.inv, momentum)
    # Scale is 0.5 of step size for half momentum update, otherwise it is just
    # the step size.
    scaled_momentum = tree_scale(scale, scaled_momentum)
    return tree_add(position, scaled_momentum)

  def _cov_scaled_noise(split, mass, tree):
    noise = random_tree(split, tree)
    noise = tensor_matmul(mass.sqrt, noise)
    return noise

  def _energy(old_momentum, new_momentum, mass, gradient, scale):
    # Accumulate the energy
    momentum_sum = tree_add(old_momentum, new_momentum)
    scaled_gradient = tensor_matmul(mass.inv, gradient)
    unscaled_energy = tree_dot(momentum_sum, scaled_gradient)
    return scale * unscaled_energy

  def _body_fun(state: LeapfrogState,
                step: jnp.array,
                parameters: schedule,
                mass: MassMatrix):
    # Full step not required in first iteration because of the half step at the
    # beginning
    positions = lax.cond(step == 0,
                         lambda pos: pos,
                         lambda pos: _position_update(
                           parameters.step_size,
                           mass,
                           pos,
                           state.momentum),
                         state.positions)

    key, split = random.split(state.key)
    noise = _cov_scaled_noise(split, mass, state.momentum)
    scaled_noise = tree_scale(
      jnp.sqrt(4 * friction * parameters.step_size),
      noise)

    data_state, mini_batch = get_data(state.data_state, information=True)
    gradient, model_state = stochastic_gradient(
      positions,
      mini_batch,
      state=state.model_state)

    decayed_momentum = tree_scale(
      1 - parameters.step_size * friction,
      state.momentum)
    negative_scaled_gradient = tree_scale(
      -1.0 * parameters.step_size,
      gradient)
    unscaled_momentum = tree_add(
      tree_add(decayed_momentum, negative_scaled_gradient),
      scaled_noise)
    updated_momentum = tree_scale(
      1 / (1 + parameters.step_size * friction),
      unscaled_momentum)

    energy = _energy(
      state.momentum,
      updated_momentum,
      mass,
      gradient,
      0.5 * parameters.step_size)
    accumulated_energy = energy + state.potential

    new_state = LeapfrogState(
      potential=accumulated_energy,
      key=key,
      positions=positions,
      momentum=updated_momentum,
      data_state=data_state,
      model_state=model_state)

    return new_state, None

  def init_fn(init_sample: PyTree,
              key: Array = None,
              batch_kwargs: Dict = None,
              init_model_state: PyTree = None,
              mass: PyTree = None):
    """Initializes the initial state of the integrator.

    Args:
      init_sample: Initial latent variables
      key: Initial PRNGKey
      batch_kwargs: Determine the initial state of the random data chain
      init_cov: Initial covariance.
      init_model_state: State of the model.

    Returns:
      Returns the initial state of the integrator.

    """

    # Initializing the initial state here makes it easier to add additional
    # variables which might be only necessary in special case
    if batch_kwargs is None:
      batch_kwargs = {}

    reference_data_state = init_data(**batch_kwargs)

    # Use constant mass if provided
    if not mass:
      if const_mass:
        mass = const_mass
      else:
        mass = init_mass(tree_util.tree_map(jnp.ones_like, init_sample))

    # Sample initial momentum
    if key is None:
      key = random.PRNGKey(0)
    key, split = random.split(key)
    momentum = _cov_scaled_noise(split, mass, init_sample)

    init_state = LeapfrogState(
      potential=jnp.array(0.0),
      key=key,
      positions=init_sample,
      momentum=momentum,
      data_state=reference_data_state,
      model_state=init_model_state)

    return init_state

  @partial(named_call, name='leapfrog_integration')
  def integrate(state: LeapfrogState,
                parameters: schedule,
                mass: PyTree = None
                ) -> LeapfrogState:

    # Use default values if mass matrix not provided
    if not mass:
      if const_mass:
        mass = const_mass
      else:
        mass = init_mass(tree_util.tree_map(jnp.ones_like, state.positions))

    # Half step for leapfrog integration
    positions = _position_update(
      0.5 * parameters.step_size, mass, state.positions, state.momentum)

    # Do the leapfrog steps
    state = LeapfrogState(
      positions=positions,
      momentum=state.momentum,
      key=state.key,
      potential=jnp.array(0.0),
      model_state=state.model_state,
      data_state=state.data_state)

    # Leapfrog integration
    state, _ = lax.scan(
      partial(_body_fun, parameters=parameters, mass=mass),
      state,
      onp.arange(steps))

    # Final half step
    positions = _position_update(
      0.5 * parameters.step_size, mass, state.positions, state.momentum)

    final_state = LeapfrogState(
      positions=positions,
      momentum=state.momentum,
      key=state.key,
      potential=state.potential,
      model_state=state.model_state,
      data_state=state.data_state)

    return final_state

  def get_fn(state: LeapfrogState) -> Dict[str, PyTree]:
    return {"variables": state.positions,
            "energy": state.potential,
            "model_state": state.model_state}

  return init_fn, integrate, get_fn


def friction_leapfrog(potential_fn: StochasticPotential,
                      batch_fn: RandomBatch,
                      steps: int = 10,
                      friction: Array = 0.25,
                      const_mass: PyTree = None,
                      noise_model = None
                      ) -> Tuple[Callable, Callable, Callable]:
  """Initializes the original SGHMC leapfrog integrator.

  Original SGHMC from [1].

  [1] https://arxiv.org/pdf/1402.4102.pdf

  Args:
    potential_fn: Likelihood and prior applied over a minibatch of data
    batch_fn: Function to draw a mini-batch of reference data
    steps: Number of intermediate leapfrog steps
    friction: Decay of momentum to counteract induced noise due to stochastic
     gradients
    const_mass: Mass matrix of the hamiltonian process
    noise_model: Stateless adaption of the noise (e. g. via the empirical fisher
      information)

  Returns:
    Returns a function running the non-conservative leapfrog integrator for T
    steps.

  """

  init_data, get_data, _ = batch_fn
  stochastic_gradient = value_and_grad(potential_fn, has_aux=True)
  if noise_model:
    init_noise_model, update_noise_model, get_noise_model = noise_model

  # Calculate the inverse and the square root
  if const_mass:
    const_mass = init_mass(const_mass)
  else:
    const_mass = None

  def _body_fun(state: LeapfrogState,
                step: jnp.array,
                parameters: schedule,
                friction: PyTree,
                mass: MassMatrix):
    del step
    scaled_momentum = tensor_matmul(mass.inv, state.momentum)
    position_update = tree_scale(parameters.step_size, scaled_momentum)
    new_positions = tree_add(state.positions, position_update)

    data_state, mini_batch = get_data(state.data_state, information=True)
    (pot, model_state), gradient = stochastic_gradient(
      new_positions,
      mini_batch,
      state=state.model_state)
    scaled_gradient = tree_scale(-parameters.step_size, gradient)
    scaled_friction = tree_scale(-parameters.step_size, friction)
    decayed_momentum = tree_multiply(scaled_friction, scaled_momentum)

    key, split = random.split(state.key)
    noise = random_tree(split, decayed_momentum)
    if noise_model:
      noise_state = update_noise_model(
        state.extra_fields,
        new_positions,
        gradient,
        friction,
        mini_batch=mini_batch,
        step_size=parameters.step_size)

      key, split = random.split(state.key)
      extra_noise = random_tree(split, decayed_momentum)

      noise_correction = get_noise_model(
        noise_state,
        new_positions,
        gradient,
        friction,
        mini_batch=mini_batch,
        step_size=parameters.step_size,
        model_state=state.model_state)
      reduced_noise = tensor_matmul(noise_correction.cb_diff_sqrt, noise)
      extra_noise = tensor_matmul(noise_correction.b_sqrt, extra_noise)
      unscaled_noise = tree_add(reduced_noise, extra_noise)
      scaled_noise = tree_scale(jnp.sqrt(2 * parameters.step_size), unscaled_noise)
    else:
      noise_state = None
      scaled_noise = tree_scale(
        jnp.sqrt(2 * friction * parameters.step_size),
        noise)

    new_momentum = tree_add(
      tree_add(decayed_momentum, scaled_gradient),
      scaled_noise)

    new_state = LeapfrogState(
      potential=pot,
      key=key,
      positions=new_positions,
      momentum=new_momentum,
      data_state=data_state,
      model_state=model_state,
      extra_fields=noise_state)

    return new_state, None

  def init_fn(init_sample: PyTree,
              key: Array = None,
              batch_kwargs: Dict = None,
              init_model_state: PyTree = None):
    """Initializes the initial state of the integrator.

    Args:
      init_sample: Initial latent variables
      key: Initial PRNGKey
      batch_kwargs: Determine the initial state of the random data chain
      init_model_state: State of the model.

    Returns:
      Returns the initial state of the integrator.

    """

    # Initializing the initial state here makes it easier to add additional
    # variables which might be only necessary in special case
    if batch_kwargs is None:
      batch_kwargs = {}

    if noise_model:
      noise_state = init_noise_model(init_sample)
    else:
      noise_state = None

    reference_data_state = init_data(**batch_kwargs)

    # Only shape of momentum is important, as momentum is resampled in each
    # integration step
    if key is None:
      key = random.PRNGKey(0)

    momentum = init_sample

    init_state = LeapfrogState(
      potential=jnp.array(0.0),
      key=key,
      positions=init_sample,
      momentum=momentum,
      data_state=reference_data_state,
      model_state=init_model_state,
      extra_fields=noise_state)

    return init_state

  @partial(named_call, name='leapfrog_integration')
  def integrate(state: LeapfrogState,
                parameters: schedule,
                mass: PyTree = None
                ) -> LeapfrogState:

    # Use default values if mass matrix not provided
    if not mass:
      if const_mass:
        mass = const_mass
      else:
        mass = init_mass(tree_util.tree_map(jnp.ones_like, state.positions))

    # If the friction is scalar
    if tree_util.treedef_is_leaf(tree_util.tree_structure(friction)):
      multiscalar_friction = tree_util.tree_map(
        partial(jnp.full_like, fill_value=friction), state.positions)
    else:
      multiscalar_friction = friction

    # Resample momentum
    key, split = random.split(state.key)
    noise = random_tree(split, state.momentum)
    momentum = tensor_matmul(mass.sqrt, noise)

    # Do the leapfrog steps
    state = LeapfrogState(
      positions=state.positions,
      momentum=momentum,
      key=key,
      potential=jnp.array(0.0),
      model_state=state.model_state,
      data_state=state.data_state,
      extra_fields=state.extra_fields)
    final_state, _ = lax.scan(
      partial(_body_fun,
              parameters=parameters,
              mass=mass,
              friction=multiscalar_friction),
      state,
      onp.arange(steps))

    return final_state

  def get_fn(state: LeapfrogState) -> Dict[str, PyTree]:
    """Returns the latent variables."""
    return {"variables": state.positions,
            "energy": state.potential,
            "model_state": state.model_state}

  return init_fn, integrate, get_fn

def langevin_diffusion(
        potential_fn: StochasticPotential,
        batch_fn: RandomBatch,
        adaption: AdaptionStrategy = None,
) -> Tuple[Callable, Callable, Callable]:
  """Initializes langevin diffusion integrator.

  Arguments:
    potential_fn: Likelihood and prior applied over a minibatch of data
    batch_fn: Function to draw a mini-batch of reference data
    adaption: Adaption of manifold for faster inference

  Returns:
    Returns a tuple consisting of ``ini_fn``, ``update_fn``, ``get_fn``.
    The init_fn takes the arguments

    - key: Initial PRNGKey
    - adaption_kwargs: Additional arguments to determine the initial manifold
      state
    - batch_kwargs: Determine the state of the random data chain

  """
  if adaption is not None:
    adapt_init, adapt_update, adapt_get = adaption
  batch_init, batch_get, _ = batch_fn
  stochastic_gradient = value_and_grad(potential_fn, argnums=0, has_aux=True)

  # We need to define an update function. All array operations must be
  # implemented via tree_map. This is probably going to change with the
  # introduction of the tree vectorizing transformation
  # --> https://github.com/google/jax/pull/3263

  # This function is intended to generate initial states. Jax key,
  # adaption, etc. can be initialized to a default value if not explicitely
  # provided

  def init_fn(init_sample: PyTree,
              key: Array = random.PRNGKey(0),
              adaption_kwargs: Dict = None,
              batch_kwargs: Dict = None,
              init_model_state: PyTree = None):
    """Initializes the state of the integrator.

    Args:
      init_sample: Initial latent variables
      key: Initial PRNGKey
      adaption_kwargs: Determine the initial state of the adaption
      batch_kwargs: Determine the initial state of the random data chain

    Returns:
      Returns the initial state of the integrator.

    """

    # Initializing the initial state here makes it easier to add additional
    # variables which might be only necessary in special case
    if adaption_kwargs is None:
      adaption_kwargs = {}
    if batch_kwargs is None:
      batch_kwargs = {}

    # Adaption is not required in the most general case
    if adaption is None:
      adaption_state = None
    else:
      adaption_state = adapt_init(init_sample, **adaption_kwargs)

    reference_data_state = batch_init(**batch_kwargs)

    init_state = LangevinState(
      key=key,
      latent_variables=init_sample,
      adapt_state=adaption_state,
      data_state=reference_data_state,
      model_state=init_model_state,
      potential=jnp.array(0.0),
      variance=jnp.array(1.0)
    )

    return init_state

  # Returns the important parameters of a state and excludes. Makes information
  # hiding possible

  def get_fn(state: LangevinState) -> Dict[str, PyTree]:
    """Returns the latent variables."""
    return {"variables": state.latent_variables,
            "likelihood": -state.potential,
            "model_state": state.model_state}

  # Update according to the integrator update rule

  @partial(named_call, name='langevin_diffusion_step')
  def update_fn(state: LangevinState,
                parameters: schedule):
    """Updates the integrator state according to a schedule.

    Args:
      state: Integrator state
      parameters: Schedule containing step_size and temperature

    Returns:
      Returns a new step calculated by applying langevin diffusion.
    """
    key, split = random.split(state.key)
    data_state, mini_batch = batch_get(state.data_state, information=True)

    noise = random_tree(split, state.latent_variables)
    (potential, (likelihoods, new_model_state)), gradient = stochastic_gradient(
      state.latent_variables,
      mini_batch,
      state=state.model_state,
      likelihoods=True)
    variance = jnp.var(likelihoods)

    scaled_gradient = tree_scale(-parameters.step_size, gradient)
    scaled_noise = tree_scale(
      jnp.sqrt(2 * parameters.temperature * parameters.step_size), noise)

    if adaption is None:
      update_step = tree_add(scaled_gradient, scaled_noise)
      adapt_state = None
    else:
      # Update the adaption
      adapt_state = adapt_update(
        state.adapt_state,
        state.latent_variables,
        gradient,
        mini_batch)
      # Get the adaption
      manifold = adapt_get(
        adapt_state,
        state.latent_variables,
        gradient,
        mini_batch)

      adapted_gradient = tensor_matmul(manifold.g_inv, scaled_gradient)
      adapted_noise = tensor_matmul(manifold.sqrt_g_inv, scaled_noise)
      scaled_gamma = tree_scale(parameters.step_size, manifold.gamma.tensor)

      update_step = tree_add(
        tree_add(scaled_gamma, adapted_gradient),
        adapted_noise)

    # Conclude the variable update by adding the step to the current samples
    new_sample = tree_add(state.latent_variables, update_step)
    new_state = LangevinState(
      key=key,
      latent_variables=new_sample,
      adapt_state=adapt_state,
      data_state=data_state,
      model_state=new_model_state,
      potential=jnp.array(potential, dtype=state.potential.dtype),
      variance=variance)

    return new_state

  return init_fn, update_fn, get_fn



from jax_sgmc.util.tree_util import (
  tree_multiply, tree_scale, tree_add,
  Array, tree_matmul, tree_dot, Tensor, tensor_matmul)
from jax_sgmc.util.list_map import (list_vmap, list_pmap, pytree_leaves_to_list,
  pytree_list_to_leaves)



# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Adapt quantities to the local or global geometry.
"""

import functools

from typing import Any, Callable, Tuple, NamedTuple

from collections import namedtuple

from jax import tree_util, flatten_util, jit, named_call, lax, vmap, grad
import jax.numpy as jnp

from jax_sgmc.util import Array, Tensor
from jax_sgmc.data import MiniBatch

PartialFn = Any
PyTree = Any

AdaptionStrategy = Callable

class AdaptionState(NamedTuple):
  """Extended adaption state returned by adaption decorator.

  This tuple stores functions to ravel and unravel the parameter and gradient
  pytree in addition to the adaption state.

  Attributes:
    state: State of the adaption strategy
    ravel_fn: Jax-partial function to transform pytree to 1D array
    unravel_fn: Jax-partial function to undo ravelling of pytree
    flat_potential: Potential function on the flattened pytree
  """
  state: PyTree
  ravel_fn: Callable
  unravel_fn: Callable
  flat_potential: Callable


class Manifold(NamedTuple):
  """Adapted manifold.

  Attributes:
    g_inv: Inverse manifold.
    sqrt_g_inv: Square root of inverse manifold.
    gamma: Diffusion to correct for positional dependence of manifold.

  """
  g_inv: Tensor
  sqrt_g_inv: Tensor
  gamma: Tensor


class MassMatrix(NamedTuple):
  """Mass matrix for HMC.

  Attributes:
    inv: Inverse of the mass matrix
    sqrt: Square root of the mass matrix

  """
  inv: PyTree
  sqrt: PyTree


class NoiseModel(NamedTuple):
  """Approximation of the gradient noise.

  Attributes:
    cb_diff_sqrt: Square root of the difference between the friction term and
      the noise model
    b_sqrt: Square root of the noise model

  """
  cb_diff_sqrt: PyTree
  b_sqrt: PyTree


def get_unravel_fn(tree: PyTree):
  """Calculates the unravel function.

  Args:
    tree: Parameter pytree

  Returns:
    Returns a jax Partial object such that the function can be passed as
    valid argument.

  """
  _, unravel_fn = flatten_util.ravel_pytree(tree)
  return tree_util.Partial(jit(unravel_fn))


def adaption(quantity: namedtuple = tuple):
  """Decorator to make adaption strategies operate on 1D arrays.

  Positional arguments are flattened while keyword arguments are passed
  unchanged.

  Args:
    quantity: Namedtuple to specify which fields are returned by
      :func:``get_adaption``.

  """
  return functools.partial(_adaption, quantity=quantity)

def _adaption(adaption_fn: Callable, quantity: namedtuple = tuple):
  @functools.wraps(adaption_fn)
  def pytree_adaption(*args, **kwargs) -> AdaptionStrategy:
    init, update, get = adaption_fn(*args, **kwargs)
    # Name call for debugging
    named_update = named_call(update, name='update_adaption_state')
    named_get = named_call(get, name='get_adapted_manifold')
    @functools.wraps(init)
    def new_init(x0: PyTree,
                 *init_args,
                 **init_kwargs) -> AdaptionState:
      # Calculate the flattened state and the ravel and unravel fun
      ravel_fn = tree_util.Partial(
        jit(lambda tree: flatten_util.ravel_pytree(tree)[0]))
      unravel_fn = get_unravel_fn(x0)

      x0_flat = ravel_fn(x0)
      init_flat = map(ravel_fn, init_args)
      state = init(x0_flat, *init_flat, **init_kwargs)

      # Wrap the potential in a flatten function if potential is provided as
      # kwarg
      minibatch_potential = kwargs.get("minibatch_potential")
      if minibatch_potential is not None:
        @tree_util.Partial
        def flat_potential(sample,
                           mini_batch,
                           model_state: PyTree = None,
                           **kwargs):
          sample_tree = unravel_fn(sample)
          if model_state is not None:
            return minibatch_potential(
              model_state, sample_tree, mini_batch, **kwargs)
          else:
            return minibatch_potential(
              sample_tree, mini_batch, **kwargs)
      else:
        flat_potential = None

      return AdaptionState(state=state,
                            ravel_fn=ravel_fn,
                            unravel_fn=unravel_fn,
                            flat_potential=flat_potential)

    @functools.wraps(update)
    def new_update(state: AdaptionState,
                   *update_args,
                   mini_batch: PyTree = None,
                   **update_kwargs):
      # Flat the params and the gradient
      flat_args = map(state.ravel_fn, update_args)

      # Update with flattened arguments
      if state.flat_potential is None:
        new_state = named_update(
          state.state,
          *flat_args,
          **update_kwargs)
      else:
        assert mini_batch, "Adaption strategy requires mini-batch"
        new_state = named_update(
          state.state,
          *flat_args,
          mini_batch,
          state.flat_potential,
          **update_kwargs)

      updated_state = AdaptionState(
        state=new_state,
        ravel_fn=state.ravel_fn,
        unravel_fn=state.unravel_fn,
        flat_potential=state.flat_potential)
      return updated_state

    @functools.wraps(get)
    def new_get(state: AdaptionState,
                *get_args,
                mini_batch: PyTree = None,
                **get_kwargs
                ) -> quantity:
      # Flat the params and the gradient
      flat_args = map(state.ravel_fn, get_args)

      # Get with flattened arguments
      if state.flat_potential is None:
        adapted_quantities = named_get(
          state.state,*flat_args, **get_kwargs)
      else:
        adapted_quantities = named_get(
          state.state, *flat_args,
          mini_batch=mini_batch, flat_potential=state.flat_potential,
          **get_kwargs)

      def unravel_quantities():
        for q in adapted_quantities:
          if q.ndim == 1:
            yield Tensor(ndim=1, tensor=state.unravel_fn(q))
          else:
            yield Tensor(ndim=2, tensor=q)

      return quantity(*unravel_quantities())
    return new_init, new_update, new_get
  return pytree_adaption


@adaption(quantity=Manifold)
def rms_prop() -> AdaptionStrategy:
  """RMSprop adaption.

  Adapt a diagonal matrix to the local curvature requiring only the stochastic
  gradient.

  Returns:
    Returns RMS-prop adaption strategy.

  [1] https://arxiv.org/abs/1512.07666
  """

  def init(sample: Array,
           alpha: Array = 0.9,
           lmbd: Array = 1e-5):
    """Initializes RMSprop algorithm.

    Args:
      sample: Initial sample to derive the sample size
      alpha: Adaption speed
      lmbd: Stabilization constant

    Returns:
      Returns the initial adaption state
    """
    v = jnp.ones_like(sample)
    return (v, alpha, lmbd)

  def update(state: Tuple[Array, Array, Array],
             sample: Array,
             sample_grad: Array,
             *args: Any,
             **kwargs: Any):
    """Updates the RMS-prop adaption.

    Args:
      state: Adaption state
      sample_grad: Stochastic gradient

    Returns:
      Returns adapted RMSprop state.
    """
    del sample, args, kwargs

    v, alpha, lmbd = state
    new_v = alpha * v + (1 - alpha) * jnp.square(sample_grad)
    return new_v, alpha, lmbd

  def get(state: Tuple[Array, Array, Array],
          sample: Array,
          sample_grad: Array,
          *args: Any,
          **kwargs: Any):
    """Calculates the current manifold of the RMS-prop adaption.

    Args:
      state: Current RMSprop adaption state

    Returns:
      Returns a manifold tuple with ``ndim == 1``.
    """
    del sample, sample_grad, args, kwargs

    v, _, lmbd = state
    g = jnp.power(lmbd + jnp.sqrt(v), -1.0)
    return g, jnp.sqrt(g), jnp.zeros_like(g)

  return init, update, get


@adaption(quantity=MassMatrix)
def mass_matrix(diagonal=True, burn_in=1000):
  """Adapt the mass matrix for HMC.

  Args:
    diagonal: Restrict the adapted matrix to be diagonal
    burn_in: Number of steps in which the matrix should be updated

  Returns:
    Returns an adaption strategy for the mass matrix.

  """

  def _update_matrix(args):
    iteration, ssq, _, _ = args
    if diagonal:
      inv = ssq / iteration
      sqrt = jnp.sqrt(iteration / ssq)
    else:
      inv = ssq / iteration
      eigw, eigv = jnp.linalg.eigh(ssq / iteration)
      # Todo: More effective computation
      sqrt = jnp.matmul(jnp.transpose(eigv), jnp.matmul(jnp.diag(jnp.sqrt(eigw)), eigv))
    return inv, sqrt

  def init(sample: Array, init_cov: Array):
    iteration = 0
    mean = jnp.zeros_like(sample)
    if diagonal:
      ssq = jnp.zeros_like(sample)
    else:
      ssq = jnp.zeros((sample.size, sample.size))

    if init_cov is None:
      init_cov = jnp.ones_like(sample)

    if diagonal:
      m_inv = init_cov
      m_sqrt = 1 / jnp.sqrt(init_cov)
    else:
      m_inv = jnp.diag(init_cov)
      m_sqrt = jnp.diag(1 / jnp.sqrt(init_cov))

    return iteration, mean, ssq, m_inv, m_sqrt

  def update(state: Tuple[Array, Array, Array, Array, Array],
             sample: Array,
             *args: Any,
             **kwargs: Any):
    del args, kwargs
    iteration, mean, ssq, m_inv, m_sqrt = state

    iteration += 1
    new_mean = (iteration - 1) / iteration * mean + 1 / iteration * sample
    if diagonal:
      ssq += jnp.multiply(sample - mean, sample - new_mean)
    else:
      ssq += jnp.outer(sample - mean, sample - new_mean)

    # Only update once
    new_m_inv, new_m_sqrt = lax.cond(
      iteration == burn_in,
      _update_matrix,
      lambda arg: (arg[2], arg[3]),
      (iteration, ssq, m_inv, m_sqrt))

    return iteration, new_mean, ssq, new_m_inv, new_m_sqrt

  def get(state: Tuple[Array, Array, Array, Array, Array]):
    _, _, _, m_inv, m_sqrt = state

    return m_inv, m_sqrt

  return init, update, get

@adaption(quantity=NoiseModel)
def fisher_information(minibatch_potential: Callable = None,
                       diagonal = True
                       ) -> AdaptionStrategy:
  """Adapts empirical fisher information.

  Use the empirical fisher information as a noise model for SGHMC. The empirical
  fisher information is approximated according to [1].

  Returns:
    Returns noise model approximation strategy.

  [1] https://arxiv.org/abs/1206.6380
  """
  assert minibatch_potential, "Fisher information requires potential function."

  def init(*args):
    del args

  def update(*args,
             **kwargs):
    del args, kwargs

  def get(state,
          sample: Array,
          sample_grad: Array,
          friction: Array,
          *args: Any,
          mini_batch: MiniBatch,
          flat_potential,
          step_size: Any = jnp.array(1.0),
          model_state: PyTree = None,
          **kwargs: Any):
    del state, args, kwargs

    _, mb_information = mini_batch
    N = mb_information.observation_count
    n = mb_information.batch_size

    # Unscale the gradient to get mean
    sample_grad /= N

    def potential_at_obs(smp, obs_idx):
      _, (likelihoods, _) = flat_potential(
        smp, mini_batch, likelihoods=True, state=model_state)
      return likelihoods[obs_idx]

    grad_diff_idx = grad(potential_at_obs, argnums=0)

    @functools.partial(vmap, out_axes=0)
    def sqd(idx):
      if diagonal:
        return jnp.square(grad_diff_idx(sample, idx) - sample_grad)
      else:
        diff = grad_diff_idx(sample, idx) - sample_grad
        return jnp.outer(diff, diff)

    ssq = jnp.sum(sqd(jnp.arange(mb_information.batch_size)), axis=0)

    if diagonal:
      v = 1 / (n - 1) * ssq
      b = 0.5 * step_size * v

      # Correct for negative eigenvalues
      correction = friction - b
      smallest_positive = jnp.min(jnp.where(correction <= 0, jnp.inf, correction))
      positive_correction = jnp.where(correction <= 0, smallest_positive, correction)

      # Apply the corrections to b
      b_corrected = friction - positive_correction

      noise_scale_sqrt = jnp.sqrt(positive_correction)
      scale_sqrt = jnp.sqrt(b_corrected)
    else:
      v = 1 / (n - 1) * ssq
      b = 0.5 * step_size * v
      eigw_cb, eigv_cb = jnp.linalg.eigh(jnp.diag(friction) - b)
      noise_scale_sqrt = jnp.matmul(
        jnp.transpose(eigv_cb),
        jnp.matmul(jnp.diag(jnp.sqrt(eigw_cb)), eigv_cb))

      eigw_b, eigv_b = jnp.linalg.eigh(b)
      # Todo: More effective computation
      scale_sqrt = jnp.matmul(
        jnp.transpose(eigv_b),
        jnp.matmul(jnp.diag(jnp.sqrt(eigw_b)), eigv_b))

    return noise_scale_sqrt, scale_sqrt

  return init, update, get



# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Solvers for Stochastic Gradient Bayesian Inference."""

from functools import partial
from typing import Callable, Any, Tuple, NamedTuple, Dict, Union

from jax import lax, jit, random, named_call
import jax.numpy as jnp
from jax_sgmc import io, util, data, integrator, potential

PyTree = Any
Array = util.Array

class AMAGOLDState(NamedTuple):
  """State of the AMAGOLD solver.

  Attributes:
    full_data_state: Cache state for full data mapping function.
    potential: True potential of the current sample
    integrator_state: State of the reversible leapfrog integrator
    key: PRNGKey for MH correction step
    acceptance_ratio: Acceptance ratio used in last step.
    mass_state: State of the mass adaption

  """
  full_data_state: data.CacheState
  potential: Array
  integrator_state: integrator.LeapfrogState
  key: Array
  acceptance_ratio: Array
  mass_state: PyTree

class SGGMCState(NamedTuple):
  """State of the AMAGOLD solver.

  Attributes:
    full_data_state: Cache state for full data mapping function.
    potential: True potential of the current sample
    integrator_state: State of the reversible leapfrog integrator
    key: PRNGKey for MH correction step
    acceptance_ratio: Acceptance ratio used in last step.
    mass_state: State of the mass adaption

  """
  full_data_state: data.CacheState
  potential: Array
  integrator_state: integrator.ObaboState
  key: Array
  acceptance_ratio: Array
  mass_state: PyTree

def mcmc(solver,
         scheduler,
         strategy="map",
         saving=None,
         loading=None):
  """Runs the solver for multiple chains and saves the collected samples

  Args:
    solver: Computes the next state form a given state.
    scheduler: Schedules solver parameters such as temperature and burn in.
    strategy: Run multiple chains in parallel or vectorized
    saving: Save samples via host_callback (if saving requires much memory)
    loading: Restore a previously saved checkpoint (scheduler state and solver
      state)

  Returns:
    Returns function which runs the solver for a given number of iterations.

  """

  if saving is None:
    # Pass the state
    init_saving, save, postprocess_saving = io.no_save()
  else:
    init_saving, save, postprocess_saving = saving

  scheduler_init, scheduler_next, scheduler_get = scheduler
  _, solver_update, solver_get = solver

  @partial(jit, static_argnums=(0,))
  def _uninitialized_update(unused_static_information, state, _):
    solver_state, scheduler_states, saving_state = state
    if len(scheduler_states) == 1:
      schedule = scheduler_get(scheduler_states[0])
      schedules = [schedule]
    else:
      schedules = util.list_vmap(scheduler_get)(*scheduler_states)
    new_state, stats = solver_update(solver_state, *schedules)
    # Kepp the sample if it is not subject to burn in or thinning.
    keep = jnp.logical_and(schedules[0].burn_in, schedules[0].accept)
    saving_state, saved = save(
      saving_state,
      keep,
      solver_get(new_state),
      scheduler_state=scheduler_states,
      solver_state=solver_state)
    # Update the scheduler with additional information from the integrator, such
    # as the acceptance ratio of AMAGOLD
    if stats is not None:
      next_fn = partial(scheduler_next, **stats)
    else:
      next_fn = scheduler_next

    if len(scheduler_states) == 1:
      scheduler_state = next_fn(scheduler_states[0])
      scheduler_states = [scheduler_state]
    else:
      scheduler_states = util.list_vmap(next_fn)(*scheduler_states)
    return (new_state, scheduler_states, saving_state), saved

  # Scan over the update function. Postprocessing the samples is required if
  # saving is None to sort out not accepted samples. Samples are not accepted
  # due tu burn in or thinning.

  # Initialize a single chain
  def _init(init_state, iterations, schedulers=None):
    main_scheduler, static_information = scheduler_init(iterations)
    # Some solvers might require additional schedulers, such as parallel
    # tempering.
    if schedulers is not None:
      helper_schedulers = [scheduler_init(iterations, **kwargs)[0]
                           for kwargs in schedulers]
      scheduler_states = [main_scheduler] + helper_schedulers
    else:
      scheduler_states = [main_scheduler]
    if loading is None:
      # Define what the saving module is expected to save
      saving_state = init_saving(
        solver_get(init_state),
        (init_state, scheduler_states),
        static_information)
    else:
      raise NotImplementedError("Loading of checkpoints is currently not "
                                "supported.")
    # Return tuple of static and non-static information
    return (iterations, static_information), (init_state, scheduler_states, saving_state)

  @partial(jit, static_argnums=0)
  def _run(static, init_state):
    iterations, static_information = static
    _update = partial(_uninitialized_update, static_information)
    (_, _, saving_state), saved = lax.scan(
      _update,
      init_state,
      jnp.arange(iterations))
    return saving_state, saved

  def run(*states: Union[PyTree, Tuple[PyTree]],
          schedulers: Union[Dict, Tuple[Dict]] = None,
          iterations: int = 1e5):
    if strategy == "map":
      def run_single():
        for state in states:
          init_args = _init(state, iterations, schedulers=schedulers)
          results = _run(*init_args)
          yield postprocess_saving(*results)
      return list(run_single())
    else:
      # Initialize the states sequentially
      static_info, init_states = zip(
        *[_init(state, iterations, schedulers=schedulers) for state in states])
      # Static information is the same for all chains
      if strategy == "pmap":
        mapped_run = util.list_pmap(partial(_run, static_info[0]))
      elif strategy == "vmap":
        mapped_run = jit(util.list_vmap(partial(_run, static_info[0])))
      else:
        raise NotImplementedError(f"Strategy {strategy} is unknown. ")
      saving_states, saved_values = zip(*mapped_run(*init_states))
      return list(map(postprocess_saving, saving_states, saved_values))
  return run


def sgmc(integrator) -> Tuple[Callable, Callable, Callable]:
  """Initializes the standard SGLD - sampler.

  This sampler simply integrates without acceptance/rejection or parallel
  tempering.

  Args:
    integrator: sgld or leapfrog.

  Returns:
    Returns a solver which depends only on external variables from the
    scheduler, such as the step size.

  """

  init_integrator, update_integrator, get_integrator = integrator

  def init(*args, **kwargs):
    return init_integrator(*args, **kwargs)

  @partial(named_call, name='update_step')
  def update(state, schedule):

    # No statistics such as acceptance ratio
    return update_integrator(state, schedule), None

  def get(state) -> Dict[str, PyTree]:
    return get_integrator(state)

  return init, update, get


def parallel_tempering(integrator,
                       sa_schedule: Callable = lambda n: 1 / n
                       ) -> Tuple[Callable, Callable, Callable]:
  """Exchange samples from a normal and tempered chain.

  This solver runs an additional tempered chain, from which no samples are
  drawn. The normal chain and the additional chain exchange samples at random
  by a reversible jump process [1].

  [1] https://arxiv.org/abs/2008.05367v3

  Args:
    integrator: standard langevin diffusion integrator
    sa_schedule: learning rate schedule to estimate the standard deviation of
      the stochastic potential

  Returns:
    Returns the reSGLD solver.

  """

  init_integrator, update_integrator, get_integrator = integrator

  # Todo: Maybe also provide pmap?

  # Todo: Currently not supported because of missing stop_vmap implementation
  # vmapped_update = util.list_vmap(lambda args: update_integrator(*args))

  def init(normal_sample: PyTree,
           tempered_sample: PyTree,
           ssq_init: Array = jnp.array(0.0),
           key: Array = random.PRNGKey(0),
           F: Array = jnp.array(1.0),
           **kwargs):
    key, split1, split2 = random.split(key, 3)

    normal_chain = init_integrator(normal_sample, key=split1, **kwargs)
    hot_chain = init_integrator(tempered_sample, key=split2, **kwargs)

    return normal_chain, hot_chain, ssq_init, F, 0, key

  # Hot schedule is an additional schedule, which has no influence on the
  # saved samples.
  @partial(named_call, name='resgld_swap_step')
  def update(state, normal_schedule, hot_schedule):
    normal_chain, hot_chain, ssq_estimate, F, step, key = state

    step += 1

    # Todo: Replace with vmapped / pmapped update
    normal_chain = update_integrator(normal_chain, normal_schedule)
    hot_chain = update_integrator(hot_chain, hot_schedule)

    # Update of std_estimate
    step_size = sa_schedule(step)
    ssq_estimate = ((1 - step_size) * ssq_estimate
                     + step_size * normal_chain.variance)

    # Swapping step
    temps = 1 / normal_schedule.temperature - 1 / hot_schedule.temperature
    correction = temps * ssq_estimate / F
    log_s =  temps * (normal_chain.potential - hot_chain.potential - correction)

    key, split = random.split(key)
    log_u = jnp.log(random.uniform(split))

    # Swap the chains at random
    (normal_chain, hot_chain) = lax.cond(
      log_u < log_s,
      lambda cold_hot: cold_hot,
      lambda cold_hot: (cold_hot[1], cold_hot[0]),
      (normal_chain, hot_chain))

    return (normal_chain, hot_chain, ssq_estimate, F, step, key), None

  # Return only the results of the normally tempered chain
  def get(state) -> Dict[str, PyTree]:
    return get_integrator(state[0])

  return init, update, get

def amagold(integrator_fn,
            full_potential_fn: potential.FullPotential,
            full_data_map: data.OrderedBatch,
            mass_adaption: Callable = None
            ) -> Tuple[Callable, Callable, Callable]:
  """Initializes AMAGOLD integration.

  Args:
    integrator_fn: Reversible leapfrog integrator.
    full_potential_fn: Function to calculate true potential.
    full_data_map: Tuple returned by :func:`jax_sgmc.data.full_reference_data``
    mass_adaption: Function to adapt a constant mass during the burn in phase.

  Returns:
    Returns the AMAGOLD solver, a combination of reversible stochastic
    hamiltonian dynamics and amortized MH corrections steps.

  """

  init_integrator, update_integrator, get_integrator = integrator_fn
  init_full_data, full_data_map_fn, _ = full_data_map
  if mass_adaption:
    init_mass, update_mass, get_mass = mass_adaption

  def init(init_sample: PyTree,
           key: Array = random.PRNGKey(0),
           initial_mass: PyTree = None,
           full_data_kwargs: dict = None,
           **kwargs) -> AMAGOLDState:
    if mass_adaption:
      mass_state = init_mass(init_sample, initial_mass)
      mass = get_mass(mass_state)
    else:
      mass_state = None
      mass = None

    if not full_data_kwargs:
      full_data_kwargs = {}
    full_data_state = init_full_data(**full_data_kwargs)

    init_model_state = kwargs.get('init_model_state')

    potential, (full_data_state, model_state) = full_potential_fn(
      init_sample,
      full_data_state,
      full_data_map_fn,
      state=init_model_state)

    kwargs['init_model_state'] = model_state

    key, split = random.split(key)
    # Todo: Init with correct covariance
    integrator_state = init_integrator(init_sample, key=key, mass=mass, **kwargs)

    state = AMAGOLDState(
      integrator_state=integrator_state,
      potential=potential,
      full_data_state=full_data_state,
      key=split,
      acceptance_ratio=(jnp.array(0.0), jnp.array(0.0)),
      mass_state=mass_state)
    return state

  @partial(named_call, name='amagold_mh_step')
  def update(state: AMAGOLDState, schedule):
    if mass_adaption:
      mass = get_mass(state.mass_state)
    else:
      mass = None

    key, split = random.split(state.key, 2)
    proposal = update_integrator(
      state.integrator_state,
      schedule,
      mass=mass)

    # MH correction step
    new_potential, (full_data_state, new_model_state) = full_potential_fn(
      proposal.positions,
      state.full_data_state,
      full_data_map_fn,
      state=proposal.model_state)

    # Limit acceptance ratio to 1.0
    log_alpha = state.potential - new_potential + proposal.potential
    log_alpha = jnp.where(log_alpha > 0.0, 0.0, log_alpha)
    slice = random.uniform(split)

    # If true, the step is accepted
    mh_integrator_state, new_potential, direction, mh_model_state = lax.cond(
      jnp.log(slice) < log_alpha,
      lambda new_old: new_old[0],
      lambda new_old: new_old[1],
      ((proposal, new_potential, 1.0, new_model_state),
       (state.integrator_state, state.potential, -1.0,
        state.integrator_state.model_state)))

    new_integrator_state = integrator.LeapfrogState(
      positions=mh_integrator_state.positions,
      momentum=util.tree_scale(direction, mh_integrator_state.momentum),
      model_state=mh_model_state,
      potential=mh_integrator_state.potential,
      # These parameters must be provided from the updated state, otherwise
      # the noise and the random data is not resampled
      data_state=proposal.data_state,
      key=proposal.key)

    # Adapt the mass on the accepted sample
    if mass_adaption:
      mass_state = update_mass(state.mass_state, mh_integrator_state.positions)
    else:
      mass_state = None

    new_state = AMAGOLDState(
      key=key,
      integrator_state=new_integrator_state,
      potential=new_potential,
      full_data_state=full_data_state,
      acceptance_ratio=(jnp.exp(log_alpha), schedule.step_size),
      mass_state=mass_state)

    stats = {'acceptance_ratio': jnp.exp(log_alpha)}

    return new_state, stats

  def get(state: AMAGOLDState) -> Dict[str, PyTree]:
    int_dict = get_integrator(state.integrator_state)
    int_dict['acceptance_ratio'] = state.acceptance_ratio[0]
    int_dict['step_size'] = state.acceptance_ratio[1]
    return int_dict

  return init, update, get

def sggmc(integrator_fn,
          full_potential_fn: potential.FullPotential,
          full_data_map: data.OrderedBatch,
          mass_adaption: Callable = None
          ) -> Tuple[Callable, Callable, Callable]:
  """Gradient Guided Monte Carlo using Stochastic Gradients.

  The OBABO integration scheme is reversible even when using stochastic
  gradients and provides second order accuracy. Therefore, a MH-acceptance step
  can be applied to sample from the correct posterior distribution.

  [1] https://arxiv.org/abs/2102.01691

  Args:
    integrator_fn: Reversible leapfrog integrator.
    full_potential_fn: Function to calculate true potential.
    mass_adaption: Function to predict the mass during warmup.

  Returns:
    Returns the SGGMC solver.

  """

  init_integrator, update_integrator, get_integrator = integrator_fn
  init_full_data, full_data_map_fn, _ = full_data_map
  if mass_adaption:
    init_mass, update_mass, get_mass = mass_adaption

  def init(init_sample: PyTree,
           key: Array = random.PRNGKey(0),
           initial_mass: PyTree = None,
           full_data_kwargs: dict = None,
           **kwargs) -> SGGMCState:

    if not full_data_kwargs:
      full_data_kwargs = {}
    full_data_state = init_full_data(**full_data_kwargs)

    init_model_state = kwargs.get('init_model_state')

    full_pot, (full_data_state, new_model_state) = full_potential_fn(
      init_sample,
      full_data_state,
      full_data_map_fn,
      state=init_model_state)

    # Pass the new model state to the integrator
    kwargs['init_model_state'] = new_model_state

    key, split = random.split(key)
    integrator_state = init_integrator(init_sample, key=key, **kwargs)

    if mass_adaption:
      mass_state = init_mass(init_sample, initial_mass)
    else:
      mass_state = None

    state = SGGMCState(
      integrator_state=integrator_state,
      potential=full_pot,
      full_data_state=full_data_state,
      key=split,
      mass_state=mass_state,
      acceptance_ratio=(jnp.array(0.0), jnp.array(0.0), jnp.array(0.0)))
    return state

  @partial(named_call, name='sggmc_mh_step')
  def update(state: SGGMCState, schedule) -> Tuple[SGGMCState, Dict]:
    # Get the current mass
    if mass_adaption:
      mass = get_mass(state.mass_state)
    else:
      mass = None

    key, split = random.split(state.key, 2)
    proposal = update_integrator(
      state.integrator_state,
      schedule,
      mass=mass)

    # MH correction step
    new_potential, (full_data_state, new_model_state) = full_potential_fn(
      proposal.positions,
      state.full_data_state,
      full_data_map_fn,
      state=proposal.model_state)

    log_alpha = - 1.0 / schedule.temperature * (
      new_potential - state.potential
      + proposal.kinetic_energy_end - proposal.kinetic_energy_start)
    # Limit the probability to 1.0 to ensure correct calculation of acceptance
    # ratio statistics.
    log_alpha = jnp.where(log_alpha <= 0, log_alpha, 0.0)

    slice = random.uniform(split)

    # If true, the step is accepted
    mh_integrator_state, new_potential, mh_model_state = lax.cond(
      jnp.log(slice) < log_alpha,
      lambda new_old: new_old[0],
      lambda new_old: new_old[1],
      ((proposal, new_potential, new_model_state),
       (state.integrator_state, state.potential,
        state.integrator_state.model_state)))

    new_integrator_state = integrator.ObaboState(
      positions=mh_integrator_state.positions,
      momentum=mh_integrator_state.momentum,
      model_state=mh_model_state,
      potential=mh_integrator_state.potential,
      # These parameters must be provided from the updated state, otherwise
      # the noise and the random data is not resampled
      data_state=proposal.data_state,
      key=proposal.key,
      kinetic_energy_start=0.0,
      kinetic_energy_end=0.0)

    # Adapt the mass
    if mass_adaption:
      mass_state = update_mass(state.mass_state, mh_integrator_state.positions)
    else:
      mass_state = None

    new_state = SGGMCState(
      key=key,
      integrator_state=new_integrator_state,
      potential=new_potential,
      full_data_state=full_data_state,
      mass_state=mass_state,
      acceptance_ratio=(jnp.exp(log_alpha), schedule.step_size, proposal.kinetic_energy_end-proposal.kinetic_energy_start))

    stats = {'acceptance_ratio': jnp.exp(log_alpha)}

    return new_state, stats

  def get(state: SGGMCState) -> Dict[str, PyTree]:
    int_dict = get_integrator(state.integrator_state)
    int_dict['acceptance_ratio'] = state.acceptance_ratio[0]
    int_dict['step_size'] = state.acceptance_ratio[1]
    int_dict['kinetic_energy'] = state.acceptance_ratio[2]
    int_dict['potential'] = state.potential
    return int_dict

  return init, update, get



# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility to evaluate stochastic or true potential.

This module transforms the likelihood function for a single observation or a
batch of observations to a function calculating the stochastic or full potential
making use of ``map``, ``vmap`` and ``pmap``.

"""

from functools import partial

from typing import Callable, Any, AnyStr, Optional, Tuple, Union, Protocol

from jax import vmap, pmap, lax, tree_util, named_call

import jax.numpy as jnp

from jax_sgmc import util
from jax_sgmc.data import CacheState, MiniBatch

PyTree = Any
Array = util.Array

Likelihood = Union[
  Callable[[PyTree, PyTree, MiniBatch], Tuple[Array, PyTree]],
  Callable[[PyTree, MiniBatch], Array]]
Prior = Callable[[PyTree], Array]

class StochasticPotential(Protocol):
  def __call__(self,
               sample: PyTree,
               reference_data: MiniBatch,
               state: PyTree = None,
               mask: Array = None,
               likelihoods: bool = False
               ) -> Union[Tuple[Array, PyTree],
                          Tuple[Array, Tuple[Array, PyTree]]]:
    """Calculates the stochastic potential for a mini-batch of data.

      Args:
        sample: Model parameters
        reference_data: Batch of observations
        state: Special parameters of the model which should not change the
          result of a model evaluation.
        mask: Marking invalid (e.g. double) samples
        likelihoods: Return the likelihoods of all model evaluations separately

      Returns:
        Returns an approximation of the true potential based on a mini-batch of
        reference data. Moreover, the likelihood for every single observation
        can be returned.

     """

class FullPotential(Protocol):
  def __call__(self,
               sample: PyTree,
               data_state: CacheState,
               full_data_map_fn: Callable,
               state: PyTree = None
               ) -> Tuple[Array, Tuple[CacheState, PyTree]]:
    """Calculates the potential over the full dataset.

      Args:
        sample: Model parameters
        data_state: State of the ``full_data_map`` functional
        full_data_map_fn: Functional mapping a function over the complete
          dataset
        state: Special parameters of the model which should not change the
          result of a model evaluation.

      Returns:
        Returns the potential of the current sample using the full dataset.

    """



# Todo: Possibly support soft-vmap (numpyro)

def minibatch_potential(prior: Prior,
                        likelihood: Callable,
                        strategy: AnyStr = "map",
                        has_state: bool = False,
                        is_batched: bool = False,
                        temperature: float = 1.) -> StochasticPotential:
  """Initializes the potential function for a minibatch of data.

  Args:
    prior: Log-prior function which is evaluated for a single
      sample.
    likelihood: Log-likelihood function. If ``has_state = True``, then the
      first argument is the model state, otherwise the arguments are ``sample,
      reference_data``.
    strategy: Determines hwo to evaluate the model function with respect for
      sample:

      - ``'map'`` sequential evaluation
      - ``'vmap'`` parallel evaluation via vectorization
      - ``'pmap'`` parallel evaluation on multiple devices

    has_state: If an additional state is provided for the model evaluation
    is_batched: If likelihood expects a batch of observations instead of a
      single observation. If the likelihood is batched, choosing the strategy
      has no influence on the computation.
    temperature: Posterior temperature. T = 1 is the Bayesian posterior.

  Returns:
    Returns a function which evaluates the stochastic potential for a mini-batch
    of data. The first argument are the latent variables and the second is the
    mini-batch.
  """

  # State is always passed to simplify usage in solvers
  def stateful_likelihood(state: PyTree,
                          sample: PyTree,
                          reference_data: PyTree):
    if has_state:
      lks, state = likelihood(state, sample, reference_data)
    else:
      lks = likelihood(sample, reference_data)
      state = None
    # Ensure that a scalar is returned to avoid broadcasting with mask
    return jnp.squeeze(lks), state

  # Define the strategies to evaluate the likelihoods sequantially, vectorized
  # or in parallel
  if is_batched:
    batched_likelihood = stateful_likelihood
  elif strategy == 'map':
    def batched_likelihood(state: PyTree,
                           sample: PyTree,
                           reference_data: PyTree):
      partial_likelihood = partial(stateful_likelihood, state, sample)
      return lax.map(partial_likelihood, reference_data)
  elif strategy == 'pmap':
    batched_likelihood = pmap(stateful_likelihood,
                              in_axes=(None, None, 0))
  elif strategy == 'vmap':
    batched_likelihood = vmap(stateful_likelihood,
                              in_axes=(None, None, 0))
  else:
    raise NotImplementedError(f"Strategy {strategy} is unknown")


  def batch_potential(state: PyTree,
                      sample: PyTree,
                      reference_data: MiniBatch,
                      mask: Array):
    # Approximate the potential by taking the average and scaling it to the
    # full data set size
    batch_data, batch_information = reference_data
    N = batch_information.observation_count
    n = batch_information.batch_size

    batch_likelihoods, new_states = batched_likelihood(
      state, sample, batch_data)
    if is_batched:
      # Batched evaluation returns single state
      new_state = new_states
    elif state is not None:
      new_state = tree_util.tree_map(
        lambda ary, org: jnp.reshape(jnp.take(ary, 0, axis=0), org.shape),
        new_states, state)
    else:
      new_state = None

    # The mask is only necessary for the full potential evaluation
    if mask is None:
      stochastic_potential = - N * jnp.mean(batch_likelihoods, axis=0)
    else:
      stochastic_potential = - N / n * jnp.dot(batch_likelihoods, mask)
    return stochastic_potential, batch_likelihoods, new_state

  @partial(named_call, name='evaluate_stochastic_potential')
  def potential_function(sample: PyTree,
                         reference_data: MiniBatch,
                         state: PyTree = None,
                         mask: Array = None,
                         likelihoods: bool = False):
    # Never differentiate w. r. t. reference data
    reference_data = lax.stop_gradient(reference_data)

    # Evaluate the likelihood and model for each reference data sample
    # likelihood_value = batched_likelihood_and_model(sample, reference_data)
    # It is also possible to combine the prior and the likelihood into a single
    # callable.

    batch_likelihood, observation_likelihoods, new_state = batch_potential(
      state, sample, reference_data, mask)

    # The prior has to be evaluated only once, therefore the extra call
    prior_value = prior(sample)

    if likelihoods:
      return (
        jnp.squeeze(batch_likelihood - prior_value) / temperature,
        (observation_likelihoods, new_state))
    else:
      return (jnp.squeeze(batch_likelihood - prior_value) / temperature,
              new_state)

  return potential_function


def full_potential(prior: Callable[[PyTree], Array],
                   likelihood: Callable[[PyTree, PyTree], Array],
                   strategy: AnyStr = "map",
                   has_state: bool = False,
                   is_batched: bool = False,
                   temperature: float = 1.,
                   ) -> FullPotential:
  """Transforms a pdf to compute the full potential over all reference data.

  Args:
    prior: Log-prior function which is evaluated for a single
      sample.
    likelihood: Log-likelihood function. If ``has_state = True``, then the
      first argument is the model state, otherwise the arguments are ``sample,
      reference_data``.
    strategy: Determines how to evaluate the model function with respect for
      sample:

      - ``'map'`` sequential evaluation
      - ``'vmap'`` parallel evaluation via vectorization
      - ``'pmap'`` parallel evaluation on multiple devices

    has_state: If an additional state is provided for the model evaluation
    is_batched: If likelihood expects a batch of observations instead of a
      single observation. If the likelihood is batched, choosing the strategy
      has no influence on the computation. In this case, the last argument of
      the likelihood should be an optional mask. The mask is an arrays with ones
      for valid observations and zeros for non-valid observations.
    temperature: Posterior temperature. T = 1 is the Bayesian posterior.

  Returns:
    Returns a function which evaluates the potential over the full dataset via
    a dataset mapping from the :mod:`jax_sgmc.data` module.

  """
  assert strategy != 'pmap', "Pmap is currently not supported"

  # Can use the potential evaluation strategy for a minibatch of data. The prior
  # must be evaluated independently.
  batch_potential = minibatch_potential(lambda _: jnp.array(0.0),
                                        likelihood,
                                        strategy=strategy,
                                        has_state=has_state,
                                        is_batched=is_batched)

  def batch_evaluation(sample, reference_data, mask, state):
    potential, state = batch_potential(sample, reference_data, state, mask)
    # We need to undo the scaling to get the real potential
    _, batch_information = reference_data
    N = batch_information.observation_count
    n = batch_information.batch_size
    unscaled_potential = potential * n / N
    return unscaled_potential, state

  @partial(named_call, name='evaluate_true_potential')
  def sum_batched_evaluations(sample: PyTree,
                              data_state: CacheState,
                              full_data_map_fn: Callable,
                              state: PyTree = None):
    body_fn = partial(batch_evaluation, sample)

    if data_state is None:  # quick fix to let it run with full_data_mapper
      results, new_state = full_data_map_fn(
        body_fn, state, masking=True, information=True)
    else:
      data_state, (results, new_state) = full_data_map_fn(
        body_fn, data_state, state, masking=True, information=True)

    # The prior needs just a single evaluation
    prior_eval = prior(sample)

    return (jnp.squeeze(jnp.sum(results) - prior_eval) / temperature,
            (data_state, new_state))

  return sum_batched_evaluations



"""Test the evaluation of the potential."""

from functools import partial
import itertools

import jax
import jax.numpy as jnp

from jax import random, jit, lax

import pytest

# from jax_sgmc.data import mini_batch
from jax_sgmc.potential import minibatch_potential, full_potential
from jax_sgmc.data import MiniBatchInformation, full_reference_data
from jax_sgmc.data.numpy_loader import NumpyDataLoader
from jax_sgmc.util import testing

# Todo: Test the potential evaluation function on arbitrary pytrees.

class TestPotential():
  # Helper functions

  @pytest.fixture
  def linear_potential(self):

    def likelihood(sample, reference_data):
      return jnp.sum(sample * reference_data)

    def prior(sample):
      return jnp.sum(sample)

    return prior, likelihood

  @pytest.fixture
  def potential(self):
    """Define likelihood with pytree sample and pytree reference data."""

    def likelihood(sample, reference_data):
      scale = sample["scale"]
      bases = sample["base"]
      powers = reference_data["power"]
      ref_scale = reference_data["scale"]
      return scale * ref_scale * jnp.sum(jnp.power(bases, powers))

    def prior(sample):
      return jnp.exp(-sample["scale"])

    return prior, likelihood

  @pytest.fixture
  def stateful_potential(self):
    def likelihood(state, sample, reference_data):
      new_state = sample
      scale = sample["scale"] * state["scale"]
      bases = sample["base"] +  state["base"]
      powers = reference_data["power"]
      ref_scale = reference_data["scale"]
      return scale * ref_scale * jnp.sum(jnp.power(bases, powers)), new_state

    def prior(sample):
      return jnp.exp(-sample["scale"])
    return prior, likelihood

  @pytest.mark.parametrize("obs, dim", itertools.product([7, 11], [3, 5]))
  def test_linear_potential(self, linear_potential, obs, dim):
    prior, likelihood = linear_potential
    # Setup potential

    scan_pot = minibatch_potential(prior, likelihood, strategy="map")
    vmap_pot = minibatch_potential(prior, likelihood, strategy="vmap")
    # pmap_pot = minibatch_potential(prior, likelihood, strategy="pmap")

    # Setup reference data
    key = random.PRNGKey(0)

    split1, split2 = random.split(key, 2)
    observations = jnp.tile(jnp.arange(4), (dim, 1))
    reference_data = observations, MiniBatchInformation(observation_count=obs,
                                                        batch_size=dim,
                                                        mask=jnp.ones(dim))
    sample = jnp.ones(4)

    true_result = -jnp.sum(jnp.arange(4)) * obs -4

    scan_result, _ = scan_pot(sample, reference_data)
    vmap_result, _ = vmap_pot(sample, reference_data)
    # pmap_result, _ = pmap_pot(sample, reference_data)

    testing.assert_close(scan_result, true_result)
    testing.assert_close(vmap_result, true_result)
    # test_util.check_close(pmap_result, true_result)

  @pytest.mark.parametrize("obs, dim", itertools.product([7, 11], [3, 5]))
  def test_stochastic_potential_zero(self, potential, obs, dim):
    _, likelihood = potential
    prior = lambda _: 0.0
    # Setup potential

    scan_pot = minibatch_potential(prior, likelihood, strategy="map")
    vmap_pot = minibatch_potential(prior, likelihood, strategy="vmap")
    # pmap_pot = minibatch_potential(prior, likelihood, strategy="pmap")

    # Setup reference data
    key = random.PRNGKey(0)

    split1, split2 = random.split(key, 2)
    observations = {"scale": random.exponential(split1, shape=(obs,)),
                    "power": random.exponential(split2, shape=(obs, dim))}
    reference_data = observations, MiniBatchInformation(observation_count=obs,
                                                        batch_size=obs,
                                                        mask=jnp.ones(dim))
    sample = {"scale": 0.5, "base": jnp.zeros(dim)}

    zero_array = jnp.array(-0.0)
    scan_result, _ = scan_pot(sample, reference_data)
    vmap_result, _ = vmap_pot(sample, reference_data)
    # pmap_result, _ = pmap_pot(sample, reference_data)

    testing.assert_equal(scan_result, zero_array)
    testing.assert_equal(vmap_result, zero_array)
    # test_util.check_close(pmap_result, zero_array)

  @pytest.mark.parametrize("obs, dim", itertools.product([7, 11], [3, 5]))
  def test_stochastic_potential_jit(self, potential, obs, dim):
    _, likelihood = potential
    prior = lambda _: 0.0
    # Setup potential

    scan_pot = jit(minibatch_potential(prior, likelihood, strategy="map"))
    vmap_pot = jit(minibatch_potential(prior, likelihood, strategy="vmap"))
    # pmap_pot = jit(minibatch_potential(prior, likelihood, strategy="pmap"))

    # Setup reference data
    key = random.PRNGKey(0)

    split1, split2 = random.split(key, 2)
    observations = {"scale": random.exponential(split1, shape=(obs,)),
                    "power": random.exponential(split2, shape=(obs, dim))}
    reference_data = observations, MiniBatchInformation(
      observation_count=obs,
      batch_size=obs,
      mask=jnp.ones(dim))
    sample = {"scale": 0.5, "base": jnp.zeros(dim)}

    zero_array = jnp.array(-0.0)
    scan_result, _ = scan_pot(sample, reference_data)
    vmap_result, _ = vmap_pot(sample, reference_data)
    # pmap_result, _ = pmap_pot(sample, reference_data)

    testing.assert_equal(scan_result, zero_array)
    testing.assert_equal(vmap_result, zero_array)
    # test_util.check_close(pmap_result, zero_array)

  @pytest.mark.parametrize("obs, dim", itertools.product([7, 11], [3, 5]))
  def test_stochastic_potential_equal(self, potential, obs, dim):
    prior, likelihood = potential
    # Setup potential

    scan_pot = jit(minibatch_potential(prior, likelihood, strategy="map"))
    vmap_pot = jit(minibatch_potential(prior, likelihood, strategy="vmap"))
    # pmap_pot = jit(minibatch_potential(prior, likelihood, strategy="pmap"))

    # Setup reference data
    key = random.PRNGKey(0)

    split1, split2, split3 = random.split(key, 3)
    observations = {"scale": random.exponential(split1, shape=(obs,)),
                    "power": random.exponential(split2, shape=(obs, dim))}
    reference_data = observations, MiniBatchInformation(
      observation_count=obs,
      batch_size=obs,
      mask=jnp.ones(dim))
    sample = {"scale": 0.5, "base": random.uniform(split3, (dim, ))}

    scan_result, _ = scan_pot(sample, reference_data)
    vmap_result, _ = vmap_pot(sample, reference_data)
    # pmap_result, _ = pmap_pot(sample, reference_data)

    testing.assert_close(scan_result, vmap_result)
    # test_util.check_close(scan_result, pmap_result)

  @pytest.mark.parametrize("obs, dim", itertools.product([7, 11], [3, 5]))
  def test_stochastic_potential_gradient_equal(self, potential, obs, dim):
    prior, likelihood = potential
    # Setup potential

    scan_grad = jit(
      jax.grad(minibatch_potential(prior, likelihood, strategy="map"),
               has_aux=True,
               argnums=0))
    vmap_grad = jit(
      jax.grad(minibatch_potential(prior, likelihood, strategy="vmap"),
               has_aux=True,
               argnums=0))
    # pmap_grad = jit(
    #   jax.grad(minibatch_potential(prior, likelihood, strategy="pmap"),
    #            has_aux=True,
    #            argnums=0))

    # Setup reference data
    key = random.PRNGKey(0)

    split1, split2, split3 = random.split(key, 3)
    observations = {"scale": random.exponential(split1, shape=(obs,)),
                    "power": random.exponential(split2, shape=(obs, dim))}
    reference_data = observations, MiniBatchInformation(
      observation_count=obs,
      batch_size=obs,
      mask=jnp.ones(dim))
    sample = {"scale": 0.5, "base": random.uniform(split3, (dim,))}

    scan_result, _ = scan_grad(sample, reference_data)
    vmap_result, _ = vmap_grad(sample, reference_data)
    # pmap_result, _ = pmap_grad(sample, reference_data)

    testing.assert_close(scan_result, vmap_result)
    # test_util.check_close(scan_result, pmap_result)

  @pytest.mark.parametrize("obs, dim", itertools.product([7, 11], [3, 5]))
  def test_stochastic_potential_gradient_shape(self, potential, obs, dim):
    _, likelihood = potential
    prior = lambda _: 0.0
    # Setup potential

    scan_grad = jit(
      jax.grad(minibatch_potential(prior, likelihood, strategy="map"),
               has_aux=True,
               argnums=0))
    vmap_grad = jit(
      jax.grad(minibatch_potential(prior, likelihood, strategy="vmap"),
               has_aux=True,
               argnums=0))
    # pmap_grad = jit(
    #   jax.grad(minibatch_potential(prior, likelihood, strategy="pmap"),
    #            has_aux=True,
    #            argnums=0))

    # Setup reference data
    key = random.PRNGKey(0)

    # Set scale to zero to get zero gradient
    split1, split2 = random.split(key, 2)
    observations = {"scale": jnp.zeros(obs),
                    "power": random.exponential(split1, shape=(obs, dim))}
    reference_data = observations, MiniBatchInformation(
      observation_count=obs,
      batch_size=obs,
      mask=jnp.ones(dim))
    sample = {"scale": 0.5, "base": random.uniform(split2, (dim,))}

    zero_gradient = jax.tree_map(jnp.zeros_like, sample)
    scan_result, _ = scan_grad(sample, reference_data)
    vmap_result, _ = vmap_grad(sample, reference_data)
    # pmap_result, _ = pmap_grad(sample, reference_data)

    print(scan_result)
    print(vmap_result)
    # print(pmap_result)

    testing.assert_equal(scan_result, zero_gradient)
    testing.assert_equal(vmap_result, zero_gradient)
    # test_util.check_close(pmap_result, zero_gradient)

  @pytest.mark.parametrize("obs, dim", itertools.product([7, 11], [3, 5]))
  def test_stateful_stochastic_potential_zero(self, stateful_potential, obs, dim):
    _, likelihood = stateful_potential
    prior = lambda _: 0.0
    # Setup potential

    scan_pot = minibatch_potential(prior, likelihood, strategy="map", has_state=True)
    vmap_pot = minibatch_potential(prior, likelihood, strategy="vmap", has_state=True)
    # pmap_pot = minibatch_potential(prior, likelihood, strategy="pmap", has_state=True)

    # Setup reference data
    key = random.PRNGKey(0)

    split1, split2 = random.split(key, 2)
    observations = {"scale": random.exponential(split1, shape=(obs,)),
                    "power": random.exponential(split2, shape=(obs, dim))}
    reference_data = observations, MiniBatchInformation(observation_count=obs,
                                                        batch_size=obs,
                                                        mask=jnp.ones(dim))
    sample = {"scale": jnp.array([0.5]), "base": jnp.ones(dim)}
    init_state = {"scale": jnp.array([0.0]), "base": jnp.zeros(dim)}

    _, new_state_map = scan_pot(sample, reference_data, state=init_state)
    _, new_state_vmap = vmap_pot(sample, reference_data, state=init_state)
    # _, new_state_pmap = pmap_pot(sample, reference_data, state=init_state)

    print(init_state)
    print(new_state_map)

    testing.assert_close(new_state_map, sample)
    testing.assert_close(new_state_vmap, sample)
    # test_util.check_close(new_state_pmap, sample)

  @pytest.mark.parametrize("obs, dim, mbsize", itertools.product([7, 11], [3, 5], [2, 3]))
  def test_full_potential(self, potential, obs, dim, mbsize):
    prior, likelihood = potential
    # Setup potential

    scan_pot = minibatch_potential(prior, likelihood, strategy="map")

    # Setup reference data
    key = random.PRNGKey(0)

    split1, split2 = random.split(key, 2)
    observations = {"scale": random.exponential(split1, shape=(obs,)),
                    "power": random.exponential(split2, shape=(obs, dim))}
    reference_data = observations, MiniBatchInformation(observation_count=obs,
                                                        batch_size=obs,
                                                        mask=jnp.ones(mbsize))
    sample = {"scale": jnp.array([0.5]), "base": jnp.ones(dim)}
    init_state = {"scale": jnp.array([0.0]), "base": jnp.zeros(dim)}

    reference_sol, _ = scan_pot(sample, reference_data, state=init_state)

    # Initialize dataloader for full potential evaluation
    data_loader = NumpyDataLoader(**observations)
    full_data_map = full_reference_data(data_loader, cached_batches_count=2, mb_size=mbsize)

    map_data_state = full_data_map[0]()
    vmap_data_state = full_data_map[0]()

    map_pot = full_potential(prior, likelihood, strategy="map")
    vmap_pot = full_potential(prior, likelihood, strategy="vmap")

    map_sol, _ = map_pot(sample, map_data_state, full_data_map[1])
    vmap_sol, _ = vmap_pot(sample, vmap_data_state, full_data_map[1])

    testing.assert_close(reference_sol, map_sol)
    testing.assert_close(reference_sol, vmap_sol)

  @pytest.mark.parametrize("obs, dim, mbsize", itertools.product([7, 11], [3, 5], [2, 3]))
  def test_variance(self, potential, obs, dim, mbsize):
    prior, likelihood = potential
    # Setup potential

    scan_pot = minibatch_potential(prior, likelihood, strategy="map")

    # Setup reference data
    key = random.PRNGKey(0)

    split1, split2 = random.split(key, 2)
    observations = {"scale": random.exponential(split1, shape=(obs,)),
                    "power": random.exponential(split2, shape=(obs, dim))}
    reference_data = observations, MiniBatchInformation(observation_count=obs,
                                                        batch_size=obs,
                                                        mask=jnp.ones(mbsize))

    sample = {"scale": jnp.array([0.5]), "base": jnp.ones(dim)}
    pot_results = lambda obs: scan_pot(
      sample,
      (jax.tree_map(partial(jnp.expand_dims, axis=0), obs),
       MiniBatchInformation(
         observation_count=1,
         batch_size=1,
         mask=jnp.ones(1))))

    likelihoods, _ = lax.map(pot_results, observations)
    true_variance = jnp.var(likelihoods)

    _, (lkls, _) = scan_pot(sample, reference_data, likelihoods=True)
    variance = jnp.var(lkls)

    testing.assert_close(variance, true_variance)


# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools

import numpy as onp

try:
  import tensorflow as tf
  import tensorflow_datasets as tfds
except ModuleNotFoundError:
  tf = None
  tfds = None

import jax
import jax.numpy as jnp
from jax import ShapeDtypeStruct

from jax import random
from jax import lax

from jax_sgmc.data.core import random_reference_data, full_reference_data, full_data_mapper
from jax_sgmc.data.numpy_loader import NumpyDataLoader, DeviceNumpyDataLoader
from jax_sgmc.util import testing

try:
  from jax_sgmc.data.tensorflow_loader import TensorflowDataLoader
except ModuleNotFoundError:
  TensorflowDataLoader = None

from jax_sgmc.util import list_vmap, list_pmap

import pytest

# Todo: Maybe parametrize size of dataset or shape
@pytest.fixture(params=[13, 20])
def dataset(request):
  obs_count = request.param

  data = {
    "ordered_indices": jnp.arange(obs_count, dtype=jnp.int_),
    "all_ones": jnp.ones(obs_count, dtype=jnp.bool_),
    "shape_5": jnp.ones((obs_count, 5), dtype=jnp.float_),
    "shape_3_5": jnp.ones((obs_count, 3, 5), dtype=jnp.int_)
  }

  data_format = {key: ShapeDtypeStruct(dtype=value.dtype, shape=value.shape)
                 for key, value in data.items()}

  return data, data_format, obs_count


class TestNumpyDataLoader:

  @pytest.fixture
  def data_loader(self, dataset):
    data, _, _ = dataset
    return NumpyDataLoader(**data)

  def test_initializer_batch_format(self, data_loader, dataset):
    _, data_format, _ = dataset
    batch_size = 7

    init_batch = data_loader.initializer_batch(batch_size)

    for key in data_format.keys():
      required_shape = (batch_size,) + data_format[key].shape[1:]
      assert required_shape == init_batch[key].shape
      assert data_format[key].dtype == init_batch[key].dtype

  @pytest.mark.parametrize("shuffle, in_epochs",
                           [(False, False),
                            (True, False),
                            (True, True)])
  def test_get_batches_random_shape(self, data_loader, dataset, shuffle, in_epochs):
    _, data_format, _ = dataset
    batch_size = 7
    cache_size = 5

    chain_id = data_loader.register_random_pipeline(cache_size=cache_size,
                                                    mb_size=batch_size,
                                                    in_epochs=in_epochs,
                                                    shuffle=shuffle)

    batch, _ = data_loader.get_batches(chain_id)

    for key in data_format.keys():
      required_shape = (cache_size, batch_size) + data_format[key].shape[1:]
      assert required_shape == batch[key].shape
      assert data_format[key].dtype == batch[key].dtype


  def test_get_batches_ordered_shape(self, data_loader, dataset):
    _, data_format, _ = dataset
    batch_size = 7
    cache_size = 5

    chain_id = data_loader.register_ordered_pipeline(cache_size=cache_size,
                                                     mb_size=batch_size)

    batch, _ = data_loader.get_batches(chain_id)

    for key in data_format.keys():
      required_shape = (cache_size, batch_size) + data_format[key].shape[1:]
      assert required_shape == batch[key].shape
      assert data_format[key].dtype == batch[key].dtype

  def test_all_keys_passed(self, dataset):
    data, _, _ = dataset

    cache_size = 3
    batch_size = 2

    # Test that all keys are returned in a batch
    data_loader = NumpyDataLoader(**data)

    chain_a = data_loader.register_random_pipeline(cache_size=cache_size,
                                                   mb_size=batch_size)
    chain_b = data_loader.register_ordered_pipeline(cache_size=cache_size,
                                                    mb_size=batch_size)

    batch_a, _ = data_loader.get_batches(chain_a)
    batch_b, _ = data_loader.get_batches(chain_b)

    assert set(batch_a.keys()) == set(data.keys())
    assert set(batch_b.keys()) == set(data.keys())

  def test_shuffeling(self, data_loader, dataset):
    _, data_format, obs_count = dataset
    batch_size = 3

    ceil_samples = int(onp.ceil(obs_count / batch_size))
    n_duplicated = ceil_samples * batch_size - obs_count

    # Get all samples in one cache and all samples in single caches
    chain_a = data_loader.register_random_pipeline(cache_size=ceil_samples,
                                                   mb_size=batch_size,
                                                   shuffle=True)
    chain_b = data_loader.register_random_pipeline(cache_size=1,
                                                   mb_size=batch_size,
                                                   shuffle=True)

    samples_a = data_loader.get_batches(chain_a)[0]["ordered_indices"]
    samples_b = onp.array([data_loader.get_batches(chain_b)[0]["ordered_indices"]
                           for _ in range(ceil_samples)])

    # There must be exactly two duplicates
    _, count_a = onp.unique(samples_a, return_counts=True)
    _, count_b = onp.unique(samples_b, return_counts=True)

    assert count_a.size == obs_count
    assert count_b.size == obs_count
    assert onp.sum(count_a == 1) == obs_count - n_duplicated
    assert onp.sum(count_b == 1) == obs_count - n_duplicated
    assert onp.sum(count_a == 2) == n_duplicated
    assert onp.sum(count_b == 2) == n_duplicated

  def test_shuffeling_in_epochs(self, data_loader, dataset):
    _, data_format, obs_count = dataset
    batch_size = 3

    ceil_samples = int(onp.ceil(obs_count / batch_size))

    # Get all samples in one cache and all samples in single caches
    chain_a = data_loader.register_random_pipeline(cache_size=ceil_samples,
                                                   mb_size=batch_size,
                                                   shuffle=True,
                                                   in_epochs=True)
    chain_b = data_loader.register_random_pipeline(cache_size=1,
                                                   mb_size=batch_size,
                                                   shuffle=True,
                                                   in_epochs=True)

    samples_a, mask_a = data_loader.get_batches(chain_a)
    samples_a = samples_a["ordered_indices"]
    returned_b = [data_loader.get_batches(chain_b)
                 for _ in range(ceil_samples)]
    samples_b = onp.array([batch["ordered_indices"] for batch, _ in returned_b])
    mask_b = onp.array([mask for _, mask in returned_b])

    # Every element must appear exactly once
    _, count_a = onp.unique(samples_a[mask_a], return_counts=True)
    _, count_b = onp.unique(samples_b[mask_b], return_counts=True)

    assert onp.sum(count_a == 1) == obs_count
    assert onp.sum(count_b == 1) == obs_count

  def test_full_data(self, data_loader, dataset):
    _, data_format, obs_count = dataset
    batch_size = 3

    ceil_samples = int(onp.ceil(obs_count / batch_size))

    # Get all samples in one cache and all samples in single caches
    chain_a = data_loader.register_ordered_pipeline(cache_size=ceil_samples,
                                                    mb_size=batch_size)
    chain_b = data_loader.register_ordered_pipeline(cache_size=1,
                                                    mb_size=batch_size)

    samples_a, mask_a = data_loader.get_batches(chain_a)
    samples_a = samples_a["ordered_indices"]
    returned_b = [data_loader.get_batches(chain_b)
                 for _ in range(ceil_samples)]
    samples_b = onp.array([batch["ordered_indices"] for batch, _ in returned_b])
    mask_b = onp.array([mask for _, mask in returned_b])

    # Every element must appear exactly once
    _, count_a = onp.unique(samples_a[mask_a], return_counts=True)
    _, count_b = onp.unique(samples_b[mask_b], return_counts=True)

    print(mask_a)
    print(samples_a)

    assert count_a.size == obs_count
    assert count_b.size == obs_count
    assert onp.sum(count_a == 1) == obs_count
    assert onp.sum(count_b == 1) == obs_count

  def test_full_data_order(self, data_loader, dataset):
    # The order of the data should not change
    _, data_format, obs_count = dataset
    batch_size = 3

    ceil_samples = int(onp.ceil(obs_count / batch_size))

    # Get all samples in one cache and all samples in single caches
    chain_a = data_loader.register_ordered_pipeline(cache_size=ceil_samples,
                                                    mb_size=batch_size)

    # Run for at least two complete iterations over the dataset
    for _ in range(3):
      cache, masks = data_loader.get_batches(chain_a)
      for it in range(batch_size):
        start_idx = it * batch_size
        expected_idx = start_idx + onp.arange(batch_size)

        indices = cache["ordered_indices"][it]
        mask = masks[it]

        # Check that expected indices correspond to actual indices for all valid
        # samples
        testing.assert_equal(indices[mask],
                           onp.squeeze(expected_idx[mask]))

  def test_seeding(self, data_loader, dataset):
    batch_size = 3

    # Check that chains with the same seed return the same batches
    chain_a1 = data_loader.register_random_pipeline(cache_size=1,
                                                   mb_size=1,
                                                   seed=1)
    chain_a2 = data_loader.register_random_pipeline(cache_size=1,
                                                    mb_size=1,
                                                    seed=1)
    chain_b = data_loader.register_random_pipeline(cache_size=1,
                                                   mb_size=1,
                                                   seed=2)

    batch_a1, _ = data_loader.get_batches(chain_a1)
    batch_a2, _ = data_loader.get_batches(chain_a2)
    batch_b, _ = data_loader.get_batches(chain_b)

    assert batch_a1["ordered_indices"] == batch_a2["ordered_indices"]
    assert batch_b["ordered_indices"] != batch_a2["ordered_indices"]


  @pytest.mark.parametrize("shuffle, in_epochs",
                           [(False, False),
                            (True, False),
                            (True, True)])
  def test_cache_size_order(self, data_loader, shuffle, in_epochs):
    """Check that changing the cache size does not influence the order. """

    cs_small = 3
    cs_big = 5
    mb_size = 2

    small_chain = data_loader.register_random_pipeline(
      cs_small, mb_size, seed=0, shuffle=shuffle, in_epochs=in_epochs)
    big_chain = data_loader.register_random_pipeline(
      cs_big, mb_size, seed=0, shuffle=shuffle, in_epochs=in_epochs)
    small_batch, _ = data_loader.get_batches(small_chain)
    big_batch, _ = data_loader.get_batches(big_chain)

    for key in small_batch.keys():
      for idx in range(cs_small):
        testing.assert_equal(small_batch[key][idx], big_batch[key][idx])

  @pytest.mark.parametrize("shuffle, in_epochs", [(False, False),
                                                  (True, False),
                                                  (True, True)])
  def test_checkpoint_random_data(self, data_loader, dataset, shuffle, in_epochs):
    _, _, obs_count = dataset
    # Check that the NumpyDataLoader can be restarted from a saved state

    cache_size = 3
    batch_size = 2

    chain_a = data_loader.register_random_pipeline(shuffle=shuffle,
                                                   in_epochs=in_epochs,
                                                   cache_size=cache_size,
                                                   mb_size=batch_size)
    chain_b = data_loader.register_random_pipeline(shuffle=shuffle,
                                                   in_epochs=in_epochs,
                                                   cache_size=cache_size,
                                                   mb_size=batch_size)

    # Run to save the state before a masked state is returned
    n_iter = int(obs_count / (cache_size * batch_size))
    for i in range(n_iter):
      data_loader.get_batches(chain_a)
      data_loader.get_batches(chain_b)

    # Checkpoint
    checkpoint_a = data_loader.save_state(chain_a)
    checkpoint_b = data_loader.save_state(chain_b)

    # Draw masked sample
    batches_a = data_loader.get_batches(chain_a)
    batches_b = data_loader.get_batches(chain_b)

    # Initialize new chains and update them to the same state
    new_chain_a = data_loader.register_random_pipeline(cache_size=cache_size,
                                                       mb_size=batch_size,
                                                       shuffle=shuffle,
                                                       in_epochs=in_epochs)
    new_chain_b = data_loader.register_random_pipeline(cache_size=cache_size,
                                                       mb_size=batch_size,
                                                       shuffle=shuffle,
                                                       in_epochs=in_epochs)

    data_loader.load_state(new_chain_a, checkpoint_a)
    data_loader.load_state(new_chain_b, checkpoint_b)

    new_batches_a = data_loader.get_batches(new_chain_a)
    new_batches_b = data_loader.get_batches(new_chain_b)

    # The old and the new chains should have returned the same samples
    testing.assert_equal(batches_a, new_batches_a)
    testing.assert_equal(batches_b, new_batches_b)

  def test_checkpoint_full_data(self, data_loader, dataset):
    _, _, obs_count = dataset
    # Check that the NumpyDataLoader can be restarted from a saved state

    cache_size = 3
    batch_size = 2

    chain_a = data_loader.register_ordered_pipeline(cache_size=cache_size,
                                                    mb_size=batch_size)
    chain_b = data_loader.register_ordered_pipeline(cache_size=cache_size,
                                                    mb_size=batch_size)

    # Run to save the state before a masked state is returned
    n_iter = int(obs_count / (cache_size * batch_size))
    for i in range(n_iter):
      data_loader.get_batches(chain_a)
      data_loader.get_batches(chain_b)

    # Checkpoint
    checkpoint_a = data_loader.save_state(chain_a)
    checkpoint_b = data_loader.save_state(chain_b)

    # Draw masked sample
    batches_a = data_loader.get_batches(chain_a)
    batches_b = data_loader.get_batches(chain_b)

    # Initialize new chains and update them to the same state
    new_chain_a = data_loader.register_ordered_pipeline(cache_size=cache_size,
                                                        mb_size=batch_size)
    new_chain_b = data_loader.register_ordered_pipeline(cache_size=cache_size,
                                                        mb_size=batch_size)

    data_loader.load_state(new_chain_a, checkpoint_a)
    data_loader.load_state(new_chain_b, checkpoint_b)

    new_batches_a = data_loader.get_batches(new_chain_a)
    new_batches_b = data_loader.get_batches(new_chain_b)

    # The old and the new chains should have returned the same samples
    testing.assert_equal(batches_a, new_batches_a)
    testing.assert_equal(batches_b, new_batches_b)

  @pytest.mark.parametrize("data",
                           [{"x": [0, 1, 2], "y": [[0, 1], [0, 2], [0, 3]]},
                            {"x": onp.arange(3), "y": onp.zeros((3, 3))},
                            {"x": jnp.arange(3), "y": jnp.zeros((3, 3))}],
                           ids=["python list", "numpy array", "device array"])
  def test_input_data_type(self, data):
    # The data returned must always be a jax device array
    data_loader = NumpyDataLoader(**data)

    chain_a = data_loader.register_random_pipeline(cache_size=1, mb_size=1)
    batch, _ = data_loader.get_batches(chain_a)

    assert type(batch["x"]) == type(jnp.array(1))
    assert type(batch["y"]) == type(jnp.array(1))

  def test_input_data_wrong_shape(self):
    # The first axis determines the total number of observations
    x = jnp.arange(5)
    y = jnp.zeros((3, 7))

    with pytest.raises(ValueError):
      NumpyDataLoader(x=x, y=y)


class TestNumpyDeviceDataLoader:

  @pytest.fixture
  def data_loader(self, dataset):
    data, _, _ = dataset
    return DeviceNumpyDataLoader(**data)

  def test_initializer_batch_format(self, data_loader, dataset):
    _, data_format, _ = dataset

    batch_size = 7

    init_batch = data_loader.initializer_batch(batch_size)

    for key in data_format.keys():
      required_shape = (batch_size,) + data_format[key].shape[1:]
      assert required_shape == init_batch[key].shape
      assert data_format[key].dtype == init_batch[key].dtype

  @pytest.mark.parametrize("data",
                           [{"x": [0, 1, 2], "y": [[0, 1], [0, 2], [0, 3]]},
                            {"x": onp.arange(3), "y": onp.zeros((3, 3))},
                            {"x": jnp.arange(3), "y": jnp.zeros((3, 3))}],
                           ids=["python list", "numpy array", "device array"])
  def test_input_data_type(self, data):
    # The data returned must always be a jax device array
    data_loader = DeviceNumpyDataLoader(**data)

    chain_a = data_loader.init_random_data()
    _, (batch, _) = data_loader.get_random_data(chain_a, batch_size=3)

    assert type(batch["x"]) == type(jnp.array(1))
    assert type(batch["y"]) == type(jnp.array(1))

  def test_input_data_wrong_shape(self):
    # The first axis determines the total number of observations
    x = jnp.arange(5)
    y = jnp.zeros((3, 7))

    with pytest.raises(ValueError):
      NumpyDataLoader(x=x, y=y)

  def test_seeding(self, data_loader):
    batch_size = 3

    # Check that chains with the same seed return the same batches
    chain_a1 = data_loader.init_random_data(key=random.PRNGKey(1))
    chain_a2 = data_loader.init_random_data(key=random.PRNGKey(1))
    chain_b = data_loader.init_random_data(key=random.PRNGKey(2))

    _, (batch_a1, _) = data_loader.get_random_data(chain_a1, batch_size)
    _, (batch_a2, _) = data_loader.get_random_data(chain_a2, batch_size)
    _, (batch_b, _) = data_loader.get_random_data(chain_b, batch_size)

    assert jnp.all(batch_a1["ordered_indices"] == batch_a2["ordered_indices"])
    assert jnp.any(batch_b["ordered_indices"] != batch_a2["ordered_indices"])

@pytest.mark.tensorflow
class TestTensorflowDataLoader:

  @pytest.fixture
  def tf_dataset(self, dataset):
    data, data_format, obs_count = dataset

    def gen():
      for idx in range(obs_count):
        yield {key: onp.array(value[idx]) for key, value in data.items()}

    output_signature = {
      key: tf.TensorSpec(shape=format.shape[1:], dtype=format.dtype)
      for key, format in data_format.items()
    }

    # Export tensorflow Dataset here to be able to test exclude keys later on
    tfds = tf.data.Dataset.from_generator(gen, output_signature=output_signature)
    return tfds

  def test_initializer_batch_format(self, tf_dataset, dataset):
    _, data_format, _ = dataset

    batch_size = 7

    data_loader = TensorflowDataLoader(tf_dataset)
    init_batch = data_loader.initializer_batch(mb_size=batch_size)

    for key in data_format.keys():
      required_shape = (batch_size,) + data_format[key].shape[1:]
      assert required_shape == init_batch[key].shape
      assert data_format[key].dtype == init_batch[key].dtype

  def test_get_batches_random_shape(self, tf_dataset, dataset):
    _, data_format, _ = dataset
    batch_size = 7
    cache_size = 5

    data_loader = TensorflowDataLoader(tf_dataset,
                                       shuffle_cache=10)

    chain_id = data_loader.register_random_pipeline(cache_size=cache_size,
                                                    mb_size=batch_size)

    batch, _ = data_loader.get_batches(chain_id)
    print(batch)

    for key in data_format.keys():
      required_shape = (cache_size, batch_size) + data_format[key].shape[1:]
      assert required_shape == batch[key].shape
      assert data_format[key].dtype == batch[key].dtype


  @pytest.mark.skip("Not Implemented")
  def test_get_batches_ordered_shape(self):
    assert False

  @pytest.mark.parametrize("num_exclude", [1, 2, 3])
  def test_exclude_keys(self, tf_dataset, dataset, num_exclude):
    _, data_format, _ = dataset

    all_keys = list(data_format.keys())
    excluded_keys = all_keys[0:num_exclude]
    contained_keys = all_keys[num_exclude:]

    data_loader = TensorflowDataLoader(tf_dataset,
                                       exclude_keys=excluded_keys)

    chain_id = data_loader.register_random_pipeline(cache_size=2,
                                                    mb_size=3)
    batch, _ = data_loader.get_batches(chain_id)

    for key in excluded_keys:
      assert key not in batch.keys()

    for key in contained_keys:
      assert key in batch.keys()

  @pytest.mark.parametrize("name, exclude_keys",
                           [("MNIST", []),
                            ("Cifar10", ["id"])])
  def test_tensorflow_datasets(self, name, exclude_keys):
    cache_size = 2
    batch_size = 3

    tf_dataset, info = tfds.load(name, split="train", with_info=True)
    data_loader = TensorflowDataLoader(tf_dataset,
                                       exclude_keys=exclude_keys)

    chain_id = data_loader.register_random_pipeline(cache_size=cache_size,
                                                    mb_size=batch_size)
    batch, _ = data_loader.get_batches(chain_id)
    init_batch = data_loader.initializer_batch()

    # Check that format and shape is correct
    for key in info.features:
      if key in exclude_keys:
        continue
      required_shape = (cache_size, batch_size) + info.features[key].shape
      required_dtype = init_batch[key].dtype
      assert required_shape == batch[key].shape
      assert required_dtype == batch[key].dtype


class TestRandomDataAccess:

  @pytest.fixture
  def data_loader(self, dataset):
    data, _, _ = dataset
    return NumpyDataLoader(**data)

  @pytest.fixture
  def example_problem(self):
    # This problem should check that each sample

    def init_test_fn(obs_count, batch_size, data_fn):
      init_data_fn, batch_fn, _ = data_fn
      num_its = int(onp.ceil(obs_count / batch_size))

      def init_fn():
        # An array to count the number of observations returned
        sample_counts = jnp.zeros(obs_count)
        data_state = init_data_fn(shuffle=True, in_epochs=True)
        return data_state, sample_counts, 0.0

      def _test_update_fn(state, _):
        data_state, sample_counts, valid_obs_count = state
        data_state, (batch, mb_info) = batch_fn(
          data_state,
          information=True
        )
        sample_counts = sample_counts.at[batch["ordered_indices"]].add(mb_info.mask)
        valid_obs_count += jnp.sum(mb_info.mask)
        return (data_state, sample_counts, valid_obs_count), None

      def test_fn(init_state):
        return lax.scan(_test_update_fn, init_state, onp.arange(num_its))

      return init_fn, test_fn

    return init_test_fn

  def test_cache_size_invalid_arguments(self, data_loader, dataset):
    _, _, obs_count = dataset
    # The cache size must be positive and should warn if the cache is bigger
    # than the total dataset.

    with pytest.raises(ValueError):
      random_reference_data(data_loader, cached_batches_count=0, mb_size=1)

    # This warning only occurs if a HostDataLoader is used
    with pytest.warns(Warning):
      random_reference_data(data_loader,
                            cached_batches_count=obs_count + 1,
                            mb_size=1)

  def test_mb_size_invalid_arguments(self, data_loader, dataset):
    _, _, obs_count = dataset
    # The batch size must be positive and smaller or equal than the total
    # observation size.

    with pytest.raises(ValueError):
      random_reference_data(data_loader, cached_batches_count=1, mb_size=0)

    with pytest.raises(ValueError):
      random_reference_data(data_loader,
                            cached_batches_count=1,
                            mb_size=obs_count + 1)


  def test_random_data_acces_with_in_epoch_shuffeling(self, data_loader, example_problem, dataset):
    _, _, obs_count = dataset

    cache_size = 2
    batch_size = 3

    data_fn = random_reference_data(data_loader, cache_size, batch_size)

    init_fn, test_fn = example_problem(obs_count, batch_size, data_fn)

    (_, test_sample_count, test_obs_size), _ = test_fn(init_fn())

    assert onp.all(test_sample_count == 1)
    assert test_obs_size == obs_count

  def test_jit_random_data(self, data_loader, example_problem, dataset):
    _, _, obs_count = dataset

    cache_size = 2
    batch_size = 3

    data_fn = random_reference_data(data_loader, cache_size, batch_size)

    init_fn, test_fn = example_problem(obs_count, batch_size, data_fn)

    test_fn_jit = jax.jit(test_fn)
    test_fn_jit(init_fn())

    (_, test_sample_count, test_obs_size), _ = test_fn_jit(init_fn())

    assert onp.all(test_sample_count == 1)
    assert test_obs_size == obs_count

  @pytest.mark.skip("Custom Batching not implemented")
  def test_vmap_random_data(self, data_loader, example_problem, dataset):
    _, _, obs_count = dataset

    vmap_size = 2

    cache_size = 2
    batch_size = 3

    data_fn = random_reference_data(data_loader, cache_size, batch_size)

    init_fn, test_fn = example_problem(obs_count, batch_size, data_fn)

    test_fn_vmap = list_vmap(test_fn)
    init_states = [init_fn() for _ in range(vmap_size)]

    results = test_fn_vmap(*init_states)

    for res in results:
      (_, test_sample_count, test_obs_size), _ = res
      assert onp.all(test_sample_count == 1)
      assert test_obs_size == obs_count

  @pytest.mark.pmap
  def test_pmap_random_data(self, data_loader, example_problem, dataset):
    _, _, obs_count = dataset

    pmap_size = jax.device_count()

    cache_size = 2
    batch_size = 3

    data_fn = random_reference_data(data_loader, cache_size, batch_size)

    init_fn, test_fn = example_problem(obs_count, batch_size, data_fn)

    test_fn_pmap = list_pmap(test_fn)
    init_states = [init_fn() for _ in range(pmap_size)]

    results = test_fn_pmap(*init_states)

    for res in results:
      (_, test_sample_count, test_obs_size), _ = res
      assert onp.all(test_sample_count == 1)
      assert test_obs_size == obs_count


class TestFullDataMapper:

  @pytest.fixture
  def data_loader(self, dataset):
    data, _, _ = dataset
    return NumpyDataLoader(**data)

  @pytest.fixture
  def example_problem_mask(self):
    def test_update_fn(batch, mask, _):
        return (jnp.sum(mask), batch["ordered_indices"], mask), None
    return test_update_fn

  @pytest.fixture
  def example_problem_no_mask(self):
    def test_update_fn(batch, _):
      return batch["ordered_indices"], None
    return test_update_fn

  @pytest.fixture
  def example_problem_pmap(self):
    @jax.pmap
    def identity_fn(x):
      return x

    def test_update_fn(batch, _):
      original_shape = jax.tree_map(jnp.shape, batch)
      pmap_batch = jax.tree_map(
        functools.partial(jnp.reshape, newshape=(jax.device_count(), -1)),
        batch)
      pmap_identity = identity_fn(pmap_batch)
      identity = jax.tree_map(jnp.reshape, pmap_identity, original_shape)
      return identity, None

    return test_update_fn

  def test_full_data_map_mask(self, dataset, data_loader, example_problem_mask):
    _, _, obs_count = dataset

    mapper, release = full_data_mapper(cached_batches_count=3,
                                       mb_size=2,
                                       data_loader=data_loader)

    results, _ = mapper(example_problem_mask, None, masking=True)
    sum_mask, samples, masks = results

    samples = onp.ravel(samples)
    masks = onp.ravel(masks)
    sum_mask = onp.sum(sum_mask)

    # Every element must appear exactly once
    _, count = onp.unique(samples[masks], return_counts=True)

    assert onp.sum(count == 1) == obs_count
    assert sum_mask == obs_count

    release()

  def test_jit_full_data_map_mask(self, dataset, data_loader,
                              example_problem_mask):
    _, _, obs_count = dataset

    mapper, release = full_data_mapper(cached_batches_count=3,
                                       mb_size=2,
                                       data_loader=data_loader)

    jit_mapped = jax.jit(lambda: mapper(example_problem_mask, None, masking=True))
    results, _ = jit_mapped()
    sum_mask, samples, masks = results

    samples = onp.ravel(samples)
    masks = onp.ravel(masks)
    sum_mask = onp.sum(sum_mask)

    # Every element must appear exactly once
    _, count = onp.unique(samples[masks], return_counts=True)

    assert onp.sum(count == 1) == obs_count
    assert sum_mask == obs_count

    release()

  def test_full_data_map_no_mask(self, dataset, data_loader, example_problem_no_mask):
    _, _, obs_count = dataset

    mapper, release = full_data_mapper(cached_batches_count=3,
                                       mb_size=2,
                                       data_loader=data_loader)

    samples, _ = mapper(example_problem_no_mask, None, masking=False)

    # Every element must appear exactly once
    _, count = onp.unique(samples, return_counts=True)

    assert onp.sum(count == 1) == obs_count

    release()

  def test_jit_full_data_map_no_mask(self, dataset, data_loader, example_problem_no_mask):
    _, _, obs_count = dataset

    mapper, release = full_data_mapper(cached_batches_count=3,
                                       mb_size=2,
                                       data_loader=data_loader)

    jit_mapped = jax.jit(lambda: mapper(example_problem_no_mask, None, masking=False))
    samples, _ = jit_mapped()

    # Every element must appear exactly once
    _, count = onp.unique(samples, return_counts=True)

    assert onp.sum(count == 1) == obs_count

    release()

  @pytest.mark.pmap
  def test_pmap_full_data_map_no_mask(self, dataset, data_loader,
                                     example_problem_pmap):
    _, _, obs_count = dataset

    mapper, release = full_data_mapper(cached_batches_count=2,
                                       mb_size=jax.device_count(),
                                       data_loader=data_loader)

    map_fn = functools.partial(mapper, example_problem_pmap)

    assert jax.device_count() != 1

    # Mapping should work when jitted
    results, _ = jax.jit(map_fn)(None)

    # Every element must appear exactly once
    _, count = onp.unique(results["ordered_indices"], return_counts=True)

    assert onp.sum(count == 1) == obs_count

    release()

  @pytest.mark.parametrize("mb_size", [1, 2])
  def test_full_data_unbatched(self, dataset, data_loader, mb_size, example_problem_no_mask):
    _, _, obs_count = dataset

    mapper, release = full_data_mapper(cached_batches_count=3,
                                       mb_size=mb_size,
                                       data_loader=data_loader)

    samples, _ = mapper(example_problem_no_mask, None, masking=False, batched=False)

    # Every element must appear exactly once
    _, count = onp.unique(samples, return_counts=True)

    assert onp.sum(count == 1) == obs_count

    release()


class TestFullDataAccess:

  @pytest.fixture
  def data_loader(self, dataset):
    data, _, _ = dataset
    return NumpyDataLoader(**data)

  @pytest.fixture
  def example_problem_mask(self):

    def init_test_fn(data_fn):
      init_data_fn, map_fn, _ = data_fn

      def init_fn():
        # An array to count the number of observations returned
        return init_data_fn()

      def _test_update_fn(batch, mask, _):
        return (jnp.sum(mask), batch["ordered_indices"], mask), None


      def test_fn(data_state):
        return map_fn(_test_update_fn, data_state, None, masking=True)

      return init_fn, test_fn
    return init_test_fn

  @pytest.fixture
  def example_problem_no_mask(self):
    def init_test_fn(data_fn):
      init_data_fn, map_fn, _ = data_fn

      def init_fn():
        # An array to count the number of observations returned
        return init_data_fn()

      def _test_update_fn(batch, _):
        return batch["ordered_indices"], None


      def test_fn(data_state):
        return map_fn(_test_update_fn, data_state, None, masking=False)

      return init_fn, test_fn
    return init_test_fn

  def test_jit_full_data_map_mask(self, dataset, data_loader,
                              example_problem_mask):
    _, _, obs_count = dataset
    data_map = full_reference_data(data_loader,
                                   cached_batches_count=3,
                                   mb_size=2)

    init_fn, test_fn = example_problem_mask(data_map)
    test_fn_jit = jax.jit(test_fn)
    test_fn_jit(init_fn())

    _, (results, _) = test_fn_jit(init_fn())
    sum_mask, samples, masks = results

    samples = onp.ravel(samples)
    masks = onp.ravel(masks)
    sum_mask = onp.sum(sum_mask)

    # Every element must appear exactly once
    _, count = onp.unique(samples[masks], return_counts=True)

    assert onp.sum(count == 1) == obs_count
    assert sum_mask == obs_count

  @pytest.mark.parametrize("cs, mb",
                           [(2, 3),
                            (3, 2),
                            (13, 2),
                            (2, 13),
                            (13, 13)])
  def test_full_data_map_no_mask(self, dataset, data_loader, example_problem_no_mask, cs, mb):
    num_maps = 3 # Check multiple iterations over the full dataset

    _, _, obs_count = dataset
    data_map = full_reference_data(data_loader,
                                   cached_batches_count=cs,
                                   mb_size=mb)

    init_fn, test_fn = example_problem_no_mask(data_map)

    data_state = init_fn()

    for it in range(num_maps):
      data_state, (samples, _) = test_fn(data_state)

      # Every element must appear exactly once and the order of the samples
      # remains unchanged
      _, count = onp.unique(samples, return_counts=True)

      assert onp.sum(count == 1) == obs_count
      assert count.size == obs_count
      testing.assert_equal(samples, onp.arange(obs_count))

  def test_jit_full_data_map_no_mask(self, dataset, data_loader, example_problem_no_mask):
    _, _, obs_count = dataset
    data_map = full_reference_data(data_loader,
                                  cached_batches_count=3,
                                   mb_size=2)

    init_fn, test_fn = example_problem_no_mask(data_map)
    test_fn_jit = jax.jit(test_fn)
    test_fn_jit(init_fn())

    _, (samples, _) = test_fn_jit(init_fn())

    # Every element must appear exactly once
    _, count = onp.unique(samples, return_counts=True)

    assert onp.sum(count == 1) == obs_count

  @pytest.mark.skip("Custom Batching not implemented")
  def test_vmap_full_data_map_no_mask(self, dataset, data_loader, example_problem_no_mask):
    pass

  @pytest.mark.pmap
  def test_pmap_full_data_map_no_mask(self, dataset, data_loader, example_problem_no_mask):
    pass

  @pytest.mark.skip("Custom Batching not implemented")
  def test_vmap_full_data_map_no_mask(self, dataset, data_loader, example_problem_no_mask):
    pass

  @pytest.mark.pmap
  def test_pmap_full_data_map_no_mask(self, dataset, data_loader, example_problem_no_mask):
    pass


# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
from jax import random
from jax import flatten_util

import pytest

from jax_sgmc import adaption
from jax_sgmc import util
from jax_sgmc.util import testing

class TestDecorator:
  """Test that the decorator transforms array adaption to tree adaption."""

  @pytest.fixture
  def random_tree(self):
    key = random.PRNGKey(0)
    split1, split2 = random.split(key)
    tree = {"a": random.normal(split1, shape=(2,)),
            "b": {"b1": jnp.array(0.0),
                  "b2": random.normal(split2, shape=(3, 4))}}

    return tree


  @pytest.mark.parametrize("test_arg, test_kwarg", [(0.0, 1.0),
                                                    (1.0, 2.0),
                                                    (1.0, 1.0)])
  def test_args_kwargs(self, random_tree, test_arg, test_kwarg):

    @adaption.adaption(adaption.Manifold)
    def init_adaption():
      def init(sample, arg, kwarg=1.0):
        return sample, arg, kwarg
      def update(state, sample, sample_grad, arg, kwarg=1.0):
        return state, sample, sample_grad, arg, kwarg
      def get(state, sample, sample_grad, arg, kwarg=1.0):
        return arg * sample, kwarg * sample_grad, sample-sample_grad
      return init, update, get

    init, update, get = init_adaption()
    init_state = init(random_tree, test_arg, kwarg=test_kwarg)
    update_state = update(init_state, random_tree, random_tree, arg=test_arg, kwarg=test_kwarg)
    manifold = get(init_state, random_tree, random_tree, arg=test_arg, kwarg=test_kwarg)

    # Assert that parameters are passed correctly

    testing.assert_equal(init_state.ravel_fn(random_tree), init_state.state[0])
    assert init_state.state[1] == test_arg
    assert init_state.state[2] == test_kwarg

    testing.assert_equal(update_state.ravel_fn(random_tree), update_state.state[1])
    testing.assert_equal(update_state.ravel_fn(random_tree), update_state.state[2])
    assert update_state.state[3] == test_arg
    assert update_state.state[4] == test_kwarg

    assert manifold.g_inv.ndim == 1
    assert manifold.sqrt_g_inv.ndim == 1
    assert manifold.gamma.ndim == 1
    testing.assert_equal(manifold.g_inv.tensor, util.tree_scale(test_arg, random_tree))
    testing.assert_equal(manifold.sqrt_g_inv.tensor, util.tree_scale(test_kwarg, random_tree))
    testing.assert_equal(manifold.gamma.tensor, util.tree_scale(0.0, random_tree))

  @pytest.mark.parametrize("diag", [True, False])
  def test_diag_full(self, diag, random_tree):
    @adaption.adaption(adaption.Manifold)
    def init_adaption():
      def init(*args):
        return None
      def update(*args):
        return None
      def get(state, sample, *args, diag=True):
        if diag:
          return jnp.ones_like(sample), jnp.ones_like(sample), jnp.ones_like(sample)
        else:
          return jnp.eye(sample.size), jnp.eye(sample.size), jnp.ones_like(sample)
      return init, update, get
    init, _, get = init_adaption()

    init_state = init(random_tree)
    manifold = get(init_state, random_tree, random_tree, None, diag=diag)

    # Assert correctness of diagonal / full manifold
    if diag:
      assert manifold.g_inv.ndim == 1
      testing.assert_equal(util.tree_multiply(manifold.g_inv.tensor, random_tree), random_tree)
    else:
      assert manifold.g_inv.ndim == 2
      testing.assert_equal(util.tree_matmul(manifold.g_inv.tensor, random_tree), random_tree)



# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import jax.numpy as jnp
from jax import random

import numpy as onp

import pytest

from jax_sgmc import scheduler
from jax_sgmc.util import testing


class TestScheduler:
  step_size = [None,
               (scheduler.polynomial_step_size,
                {"a": 1.0, "b": 1.0, "gamma": 0.55}),
               (scheduler.polynomial_step_size_first_last,
                {"first": 1.0, "last": 0.1, "gamma": 0.35})]

  burn_in = [None,
             (scheduler.initial_burn_in,
              {"n": 100})]

  temperature = [None,
                 (scheduler.constant_temperature,
                  {"tau": 0.5})]

  @pytest.mark.parametrize(
    "step_size, burn_in, temperature",
    itertools.product(step_size, burn_in, temperature))
  def test_scheduler(self, step_size, burn_in, temperature):
    """Test that the scheduler can initialize all specific schedulers. Moreover,
    the capability to provide default values is tested. """

    # Initialize all the specific schedulers
    if step_size is not None:
      fun, kwargs = step_size
      step_size = fun(**kwargs)
    if burn_in is not None:
      fun, kwargs = burn_in
      burn_in = fun(**kwargs)
    if temperature is not None:
      fun, kwargs = temperature
      temperature = fun(**kwargs)

    # Initialize the specific scheduler

    schedule = scheduler.init_scheduler(step_size=step_size,
                                        burn_in=burn_in,
                                        temperature=temperature)

    iterations = 100
    state, _ = schedule[0](iterations)

    for _ in range(iterations):

      sched = schedule[2](state)
      state = schedule[1](state)

      assert sched.step_size.shape == tuple()
      assert sched.temperature.shape == tuple()
      assert sched.burn_in.shape == tuple()
      assert sched.accept


class TestStepSize():

  @pytest.mark.parametrize("first, last, gamma, iterations",
                          itertools.product([1.0, 0.05],
                                            [0.01, 0.0009],
                                            [0.33, 0.55],
                                            [100, 14723]))
  def test_first_last(self, first, last, gamma, iterations):
    """Test, that the first and last step size are computed right."""

    first = jnp.array(first)
    last = jnp.array(last)
    gamma = jnp.array(gamma)

    schedule = scheduler.polynomial_step_size_first_last(first=first,
                                                         last=last,
                                                         gamma=gamma)

    state = schedule.init(iterations)

    testing.assert_close(schedule.get(state, 0), first)
    testing.assert_close(schedule.get(state, iterations-1), last)


class TestBurnIn():

  @pytest.mark.parametrize("n", [123, 243])
  def test_initial_burn_in(self, n):
    """Test, that no off by one error exists."""
    burn_in = scheduler.initial_burn_in(n=n)

    state, _ = burn_in.init(1000)

    # Check that samples are not accepted
    for idx in range(n):
      bi = burn_in.get(state, idx)
      state = burn_in.update(state, idx)
      assert bi == 0.0

    # Check that next sample is accepted
    assert burn_in.get(state, n + 1) == 1.0

class TestThinning():

  @pytest.fixture(params=[100, 1000])
  def burn_in(self, request):
    iterations = request.param
    accepted = random.bernoulli(random.PRNGKey(0), p=0.3, shape=(100,))
    init = lambda *args: (None, onp.sum(accepted))
    update = lambda *args, **kwargs: None
    get = lambda _, iteration, **kwargs: accepted[iteration]
    nonzero, = onp.nonzero(accepted)
    return scheduler.specific_scheduler(init, update, get), nonzero, iterations

  def test_random_thinning(self, burn_in):
    """Given a burn in schedule, the sample can only be accepted if it is not
    subject to burn in."""

    burn_in, non_zero, iterations = burn_in
    step_size = scheduler.polynomial_step_size(a=1.0, b=1.0, gamma=1.0)

    thinning = scheduler.random_thinning(
      step_size_schedule=step_size,
      burn_in_schedule=burn_in,
      selections=int(0.5 * non_zero.size))
    state, _ = thinning.init(iterations)

    accepted = 0
    for idx in range(iterations):
      # If the state is accepted, it must also be not subject to burn in
      if thinning.get(state, idx):
        assert (idx in non_zero)
        accepted += 1
    assert accepted == int(0.5 * non_zero.size)


# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the convergence of the solver on small toy problem. """

import jax
from jax import random
import jax.numpy as jnp
from jax.scipy.stats import norm

import pytest

from jax_sgmc import data
from jax_sgmc import potential
from jax_sgmc import alias
from jax_sgmc.data.numpy_loader import NumpyDataLoader

@pytest.fixture
def problem():

  # Reference Data

  N = 4
  samples = 1000  # Total samples
  sigma = 0.5

  key = random.PRNGKey(0)
  split1, split2, split3 = random.split(key, 3)

  w = random.uniform(split3, minval=-1, maxval=1, shape=(N, 1))
  noise = sigma * random.normal(split2, shape=(samples, 1))
  x = random.uniform(split1, minval=-10, maxval=10, shape=(samples, N))
  x = jnp.stack([x[:, 0] + x[:, 1], x[:, 1], 0.1 * x[:, 2] - 0.5 * x[:, 3],
                 x[:, 3]]).transpose()
  y = jnp.matmul(x, w) + noise
  w_init = sample = {"w": jnp.zeros((N, 1)), "sigma": jnp.array(10.0)}

  M = 10
  cs = 1000

  data_loader = NumpyDataLoader(x=x, y=y)
  batch_fn = data.random_reference_data(data_loader,
                                        cached_batches_count=cs,
                                        mb_size=M)

  def model(sample, observations):
    weights = sample["w"]
    predictors = observations["x"]
    return jnp.dot(predictors, weights)

  def likelihood(sample, observations):
    sigma = sample["sigma"]
    y = observations["y"]
    y_pred = model(sample, observations)
    return norm.logpdf(y - y_pred, scale=sigma)

  def prior(unused_sample):
    return 0.0

  # If the model is more complex, the strategy can be set to map for sequential
  # evaluation and pmap for parallel evaluation.
  potential_fn = potential.minibatch_potential(prior=prior,
                                               likelihood=likelihood,
                                               strategy="vmap")
  full_potential_fn = potential.full_potential(prior=prior,
                                               likelihood=likelihood,
                                               strategy="vmap")
  return data_loader, batch_fn, potential_fn, full_potential_fn, w, w_init

class TestAlias:

  @pytest.mark.solver
  def test_rms_prop(self, problem):
    data_loader, batch_fn, potential_fn, _, w, w_init = problem

    solver = alias.sgld(
      potential_fn,
      data_loader,
      cache_size=512,
      batch_size=10,
      first_step_size=0.05,
      last_step_size=0.001,
      burn_in=20000,
      accepted_samples=4000,
      rms_prop=True
    )

    results = solver(w_init, iterations=50000)

    # Check that the standard deviation is close
    assert jnp.all(
      jnp.abs(results[0]["samples"]["variables"]["sigma"] - 0.5) < 0.5)

  @pytest.mark.solver
  def test_re_sgld(self, problem):
    data_loader, batch_fn, potential_fn, _, w, w_init = problem

    solver = alias.re_sgld(
      potential_fn,
      data_loader,
      cache_size=512,
      batch_size=10,
      first_step_size=0.0001,
      last_step_size=0.000005,
      burn_in=20000,
      accepted_samples=4000,
      temperature=100.0
    )

    results = solver((w_init, w_init), iterations=50000)

    # Check that the standard deviation is close
    assert jnp.all(
      jnp.abs(results[0]["samples"]["variables"]["sigma"] - 0.7) < 0.7)

  @pytest.mark.solver
  def test_amagold(self, problem):
    data_loader, batch_fn, potential_fn, full_potential_fn, w, w_init = problem

    solver = alias.amagold(
      potential_fn,
      full_potential_fn,
      data_loader,
      cache_size=512,
      batch_size=64,
      first_step_size=0.005,
      last_step_size=0.0005,
      burn_in=2000
    )

    results = solver(w_init, iterations=50000)

    # Check that the standard deviation is close
    assert jnp.all(
      jnp.abs(results[0]["samples"]["variables"]["sigma"] - 0.5) < 0.5)

  @pytest.mark.solver
  def test_sggmc(self, problem):
    data_loader, batch_fn, potential_fn, full_potential_fn, w, w_init = problem

    solver = alias.sggmc(
      potential_fn,
      full_potential_fn,
      data_loader,
      cache_size=512,
      batch_size=64,
      first_step_size=0.005,
      last_step_size=0.0005,
      burn_in=2000
    )

    results = solver(w_init, iterations=50000)

    # Check that the standard deviation is close
    assert jnp.all(
      jnp.abs(results[0]["samples"]["variables"]["sigma"] - 0.5) < 0.5)

  @pytest.mark.solver
  def test_obabo(self, problem):
    data_loader, batch_fn, potential_fn, _, w, w_init = problem

    solver = alias.obabo(
      potential_fn,
      data_loader,
      cache_size=512,
      batch_size=10,
      first_step_size=0.05,
      last_step_size=0.001,
      friction=1000,
      burn_in=2000
    )

    results = solver(w_init, iterations=50000)

    # Check that the standard deviation is close
    assert jnp.all(
      jnp.abs(results[0]["samples"]["variables"]["sigma"] - 0.5) < 0.5)



# Copyright 2021 Multiscale Modeling of Fluid Materials, TU Munich
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

import numpy as onp

import jax.numpy as jnp
from jax import tree_util

from jax import Array

@tree_util.register_pytree_node_class
class JaxUUID:

  def __init__(self, ints: Array = None):
    if ints is None:
      uuid_int = uuid.uuid4().int
      ints = [(uuid_int >> bits) & 0xFFFFFFFF for bits in range(0, 128, 32)]
      ints = onp.array(ints, dtype=jnp.int32)

    self._uuid_int = ints

  @property
  def as_uuid(self):
    # Rearrange 4 int32 to uuid
    shifted_ints = [int(int(sint) << bits)
                    for sint, bits
                    in zip(self._uuid_int, range(0, 128, 32))]
    int128 = sum(shifted_ints)
    # Ensure that the hex number has exactly 128 bits and is non-negative
    hex128 = hex(int128).replace('0x', '').replace('-','').zfill(32)
    return uuid.UUID(hex128)

  def __repr__(self):
    return str(self.as_uuid)

  @property
  def as_int32s(self):
    return self._uuid_int

  def tree_flatten(self):
    # Wrapping the ints in a tuple ensures that they remain a single array of
    # length 4.
    children = (self._uuid_int,)
    return (children, None)

  @classmethod
  def tree_unflatten(cls, _, children):
    ints, = children
    return cls(ints=ints)
